# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of stochastic modules."""

import abc
import dataclasses
import enum
import logging
from typing import Any, Callable, Optional, Sequence, TypeVar, Union
import zlib

from dinosaur import coordinate_systems
from dinosaur import typing
import gin
import haiku as hk
import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp


tfb = tfp.bijectors
tree_map = jax.tree_util.tree_map
tree_leaves = jax.tree_util.tree_leaves

Numeric = typing.Numeric
Quantity = typing.Quantity
_SOFTPLUS_INVERSE_1 = 0.5413248546129181

# CoreRandomState is advanced by a RandomField, and .to_*_values(core_state)
# produces the final (usable) random Array.
CoreRandomState = typing.Pytree
RandomnessState = typing.RandomnessState


def _validate_randomness_state(state: RandomnessState) -> None:
  """Validates that `state.core` is not `None`, raises an error otherwise."""
  if state.core is None:
    raise ValueError(
        f'Got {state.core=} when value is expected. '
        'Check how incoming randomness is initialized.'
    )


def make_positive_scalar(raw_parameter: typing.Array) -> jax.Array:
  """Positive [batch] scalar values, maps 0 --> 1 using a softplus(...)."""
  raw_parameter = jnp.asarray(raw_parameter)
  return jax.nn.softplus(raw_parameter + _SOFTPLUS_INVERSE_1)


# pylint: disable=logging-fstring-interpolation


################################################################################
# Single random fields that stand on their own.
################################################################################


class PreferredRepresentation(enum.Enum):
  """The preferred (for computational reasons) representation of a field."""

  NODAL = 'NODAL'
  MODAL = 'MODAL'


class RandomField(abc.ABC):
  """Base class for random fields."""

  def __init__(self, coords):
    self.coords = coords

  @property
  @abc.abstractmethod
  def preferred_representation(self) -> PreferredRepresentation | None:
    """The PreferredRepresentation for this field, or None if no preference."""

  @abc.abstractmethod
  def unconditional_sample(self, rng: typing.PRNGKeyArray) -> RandomnessState:
    """Sample the random field unconditionally."""

  @abc.abstractmethod
  def advance(self, state: RandomnessState) -> RandomnessState:
    """Updates the core state of a random field."""

  @abc.abstractmethod
  def to_modal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Returns the modal rep. of the random field specified by this class."""

  @abc.abstractmethod
  def to_nodal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Returns the nodal rep. of the random field specified by this class."""


RandomnessModule = Callable[..., RandomField]


_ADVANCE_SALT = zlib.crc32(b'advance')  # arbitrary uint32 value


T = TypeVar('T', typing.PRNGKeyArray, None)


def _prng_key_for_current_advance_step(
    randomness: typing.RandomnessState,
) -> typing.PRNGKeyArray | None:
  """Get a PRNG Key suitable for randomness in the current advance step."""
  if randomness.prng_key is None:
    return None
  salt = jnp.uint32(_ADVANCE_SALT) + jnp.uint32(randomness.prng_step)
  return jax.random.fold_in(randomness.prng_key, salt)


@gin.register
class NoRandomField(RandomField):
  """Module that disables randomness in a given module returning `None`."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      prefer_nodal: bool = True,
  ):
    """Constructs a ZerosRandomField.

    Args:
      coords: horizontal and vertical grid data.
      dt: nondimensionalized model time step.
      physics_specs: physical constants and definition of custom units.
      aux_features: additional static data.
      prefer_nodal: Whether this field should prefer a nodal representation.
    """
    super().__init__(coords)
    logging.info('[NGCM] Initializing NoRandomField')
    del dt, physics_specs, aux_features, prefer_nodal  # unused.

  @property
  def preferred_representation(self) -> PreferredRepresentation | None:
    return None

  def unconditional_sample(
      self, rng: typing.PRNGKeyArray | None
  ) -> RandomnessState:
    """Returns a zeros initialized state."""
    return RandomnessState(prng_key=rng, prng_step=0)

  def advance(self, state: RandomnessState) -> RandomnessState:
    """Updates the state of a random gaussian field."""
    return RandomnessState(
        prng_key=state.prng_key, prng_step=state.prng_step + 1
    )

  def to_nodal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    del core_state  # unused.
    return None

  def to_modal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    del core_state  # unused.
    return None


@gin.register
class ZerosRandomField(RandomField):
  """Implements a constant random field identically equal to zero."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      prefer_nodal: bool = True,
  ):
    """Constructs a ZerosRandomField.

    Args:
      coords: horizontal and vertical grid data.
      dt: nondimensionalized model time step.
      physics_specs: physical constants and definition of custom units.
      aux_features: additional static data.
      prefer_nodal: Whether this field should prefer a nodal representation.
    """
    super().__init__(coords)
    logging.info('[NGCM] Initializing ZerosRandomField')
    del dt  # unused
    del physics_specs  # unused.
    del aux_features  # unused.
    self._prefer_nodal = prefer_nodal

  @property
  def preferred_representation(self) -> PreferredRepresentation | None:
    if self._prefer_nodal:
      return PreferredRepresentation.NODAL
    else:
      return PreferredRepresentation.MODAL

  def unconditional_sample(
      self, rng: typing.PRNGKeyArray | None
  ) -> RandomnessState:
    """Returns a zeros initialized state."""
    if self._prefer_nodal:
      core = jnp.zeros(self.coords.horizontal.nodal_shape)
    else:
      core = jnp.zeros(self.coords.horizontal.modal_shape)
    return RandomnessState(
        core=core,
        nodal_value=jnp.zeros(self.coords.horizontal.nodal_shape),
        modal_value=jnp.zeros(self.coords.horizontal.modal_shape),
        prng_key=rng,
        prng_step=0,
    )

  def advance(self, state: RandomnessState) -> RandomnessState:
    """Updates the state of a random gaussian field."""
    _validate_randomness_state(state)
    return RandomnessState(
        core=jnp.zeros_like(state.core),
        nodal_value=jnp.zeros(self.coords.horizontal.nodal_shape),
        modal_value=jnp.zeros(self.coords.horizontal.modal_shape),
        prng_key=state.prng_key,
        prng_step=state.prng_step + 1,
    )

  def to_nodal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Returns the ready-for-use Zeros random field."""
    return jnp.zeros(self.coords.horizontal.nodal_shape)

  def to_modal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Returns the ready-for-use Zeros random field."""
    return jnp.zeros(self.coords.horizontal.modal_shape)


@gin.register
class GaussianRandomField(RandomField):
  """Implements gaussian random field with spatial and temporal correlations.

  This type of random fields is used in SPPT (stochastic physics
  parameterization tendencies) schemes, where each tendency due to physics
  parameterizations are multiplicatively perturbed by the value of such field.

  For implementation details see Appendix 8 in http://shortn/_56HCcQwmSS.

  With x ∈ EarthSurface, this field U is initialized at t=0 with
    U(0, x) = Σₖ Ψₖ(x) (1 - φ²)^(-0.5) σₖ γₖ σₖ ηₖ₀,
  where Ψₖ is the kth spherical harmonic basis function, φ² is the one timestep
  correlation, σₖ > 0 is a scaling factor, and ηₖ₀ are iid 1D unit Gaussians.

  With `variance` an init kwarg,
    E[U(0, x)] ≡ 0,
    1 / (4πR²) ∫ Var(U(0, x))dx = variance,
  regardless of coords (and the radius).

  Further states are generated with the recursion
    U(t + δ) = ϕ U(t) + σₖ ηₖₜ
  This ensures that U is stationary.

  In general,
    Cov(U(t, x), U(t + δ, y)) = ϕᵟ Σₖ Ψₖ(x) Ψₖ(y) (γₖ)².
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      correlation_time: Union[jax.Array, Quantity, str] = gin.REQUIRED,
      correlation_length: Union[jax.Array, Quantity, str] = gin.REQUIRED,
      variance: Optional[Union[jax.Array, Quantity, str]] = gin.REQUIRED,
      clip: float = 6.0,
  ):
    """Constructs a GaussianRandomField.

    Args:
      coords: horizontal and vertical grid data.
      dt: nondimensionalized model time step.
      physics_specs: physical constants and definition of custom units.
      aux_features: additional static data.
      correlation_time: timescale with units over which autoregressive process
        decorrelates. Typical values in NWP range from hours to days.
      correlation_length: lengthscale with units over which random field is
        correlated. Typical values in NWP range from 500-2500 km.
      variance: The average (over EarthSurface) variance of the random field If
        None, this GRF always returns a zeros field and no RNGS are drawn.
      clip: number of standard deviations at which to clip randomness to ensure
        numerical stability.
    """
    del aux_features  # unused.
    super().__init__(coords)
    logging.info(
        '[NGCM] Initializing GaussianRandomField (possibly via'
        f' CenteredLognormalRandomField) with {variance=}, {correlation_time=},'
        f' {correlation_length=}'
    )

    tau = maybe_nondimensionalize(correlation_time, physics_specs)
    correlation_length = maybe_nondimensionalize(
        correlation_length, physics_specs
    )

    # In sampling, phi appears as 1 - phi**2 = 1 - exp(-2 dt / tau)
    self.one_minus_phi2 = -jnp.expm1(-2 * dt / tau)

    self.phi = jnp.exp(-dt / tau)

    self._variance = maybe_nondimensionalize(variance, physics_specs)  # σ²

    # [Palmer] states correlation_length = sqrt(2κT) / R, therefore
    self.kt = (correlation_length / self.coords.horizontal.radius) ** 2 / 2
    self.clip = clip

  @property
  def preferred_representation(self) -> PreferredRepresentation | None:
    return PreferredRepresentation.MODAL

  @property
  def _surf_area(self) -> jax.Array:
    """Surface area of sphere of radius self.coords.horizontal.radius."""
    return 4 * jnp.pi * self.coords.horizontal.radius**2  # pytype: disable=bad-return-type  # jnp-type

  def _sigma_array(self) -> jax.Array:
    """Array of σₙ from Appendix 8 in [Palmer] http://shortn/_56HCcQwmSS."""
    # n = [0, 1, ..., N]
    n = self.coords.horizontal.modal_axes[1]  # total wavenumbers.

    # Number of longitudinal wavenumbers at each total wavenumber n.
    #  L = 2n + 1, except for the last entry.
    n_longitudian_wavenumbers = self.coords.horizontal.mask.sum(axis=0)

    # sigmas_unnormed[n] is proportional to the standard deviation for each
    # longitudinal wavenumbers at each total wavenumber n.
    sigmas_unnormed = jnp.exp(-0.5 * self.kt * n * (n + 1))

    # The sum of unnormalized variance for all longitudinal wavenumbers at each
    # total wavenumber.
    sum_unnormed_vars = jnp.sum(n_longitudian_wavenumbers * sigmas_unnormed**2)

    # This is analogous to F₀ from [Palmer].
    # (normalization * sigmas_unnormed)² would sum to 1. The leading factor
    #   self._integrated_grf_variance * (1 - self.phi ** 2)
    # ensures that the AR(1) process has variance self._integrated_grf_variance.
    # We do not include the extra fator of 2 in the denominator. I do not know
    # why [Palmer] has this factor.
    normalization = jnp.sqrt(
        self._integrated_grf_variance()
        * self.one_minus_phi2
        / sum_unnormed_vars
    )

    # The factor of coords.horizontal.radius appears because our basis vectors
    # have L2 norm = radius.  See http://screen/9FYVXZ5cMHoGDZk
    return normalization * sigmas_unnormed / self.coords.horizontal.radius

  def unconditional_sample(self, rng: typing.PRNGKeyArray) -> RandomnessState:
    """Returns a randomly initialized state for the autoregressive process."""
    modal_shape = self.coords.horizontal.modal_shape
    rng, next_rng = jax.random.split(rng)
    if self.variance is None:
      return RandomnessState(
          core=jnp.zeros(modal_shape),
          nodal_value=jnp.zeros(self.coords.horizontal.nodal_shape),
          modal_value=jnp.zeros(modal_shape),
          prng_key=next_rng,
          prng_step=0,
      )
    sigmas = self._sigma_array()
    weights = jnp.where(
        self.coords.horizontal.mask,
        jax.random.truncated_normal(rng, -self.clip, self.clip, modal_shape),
        jnp.zeros(modal_shape),
    )
    core = self.one_minus_phi2 ** (-0.5) * sigmas * weights
    return RandomnessState(
        core=core,
        nodal_value=self.to_nodal_values(core),
        modal_value=self.to_modal_values(core),
        prng_key=next_rng,
        prng_step=0,
    )

  def advance(self, state: RandomnessState) -> RandomnessState:
    """Updates the CoreRandomState of a random gaussian field."""
    _validate_randomness_state(state)
    if self.variance is None:
      return RandomnessState(
          core=jnp.zeros_like(state.core),
          nodal_value=jnp.zeros(self.coords.horizontal.nodal_shape),
          modal_value=jnp.zeros(self.coords.horizontal.modal_shape),
          prng_key=state.prng_key,
          prng_step=state.prng_step + 1,
      )
    modal_shape = self.coords.horizontal.modal_shape
    rng = _prng_key_for_current_advance_step(state)
    eta = jax.random.truncated_normal(rng, -self.clip, self.clip, modal_shape)
    next_core = state.core * self.phi + self._sigma_array() * jnp.where(
        self.coords.horizontal.mask, eta, jnp.zeros(modal_shape)
    )
    return RandomnessState(
        core=next_core,
        nodal_value=self.to_nodal_values(next_core),
        modal_value=self.to_modal_values(next_core),
        prng_key=state.prng_key,
        prng_step=state.prng_step + 1,
    )

  @property
  def variance(self) -> Numeric | None:
    """An estimate of pointwise (in nodal space) variance of this random field.

    This random field is defined in spectral space, and has no precise
    pointwise variance quantity. However, it does have a precise integrated
    variance, which is used to define the field.

    If we assume the field is stationary (with higher spectral
    precision it is near stationary), then the average of this quantity is a
    good pointwise estimate. So define
      σ² := (1 / (4πR²)) ∫ Var(U(0, x))dx
          = (1 / (4πR²)) integrated_grf_variance

    Therefore the init parameter `variance` can be used to define
      `_integrated_grf_variance := variance * surf_area`
    and then `_integrated_grf_variance` is used to define this field. The result
    is a field with pointwise variance close to the init kwarg `variance`.

    Returns:
      Numeric estimate of pointwise variance.
    """
    return self._variance

  def _integrated_grf_variance(self) -> Numeric | None:
    """Integral of the GRF's variance over the earth's surface."""
    if self.variance is None:
      return self.variance
    return self.variance * self._surf_area

  def to_modal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Returns the ready-for-use Gaussian random field."""
    return core_state

  def to_nodal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Returns the ready-for-use Gaussian random field."""
    return self.coords.horizontal.to_nodal(core_state)


@gin.register
class GaussianRandomFieldModule(GaussianRandomField, hk.Module):
  """Module wrapper of GaussianRandomField with trainable parameters."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      initial_correlation_time: Union[Quantity, str] = gin.REQUIRED,
      initial_correlation_length: Union[Quantity, str] = gin.REQUIRED,
      initial_variance: Optional[Union[Quantity, str]] = gin.REQUIRED,
      variance_bound: Optional[Union[Quantity, str]] = gin.REQUIRED,
      tune_variance: bool = True,
      clip: float = 6.0,
      name: Optional[str] = None,
  ):
    """Constructs a GaussianRandomFieldModule.

    Stochastic parameters are initialized at provided `initial_*` values.
    This hk.Module can then be used to tune values.

    Args:
      coords: horizontal and vertical grid data.
      dt: nondimensionalized model time step.
      physics_specs: physical constants and definition of custom units.
      aux_features: additional static data.
      initial_correlation_time: timescale with units over which autoregressive
        process decorrelates. Typical values in NWP range from hours to days.
      initial_correlation_length: lengthscale with units over which random field
        is correlated. Typical values in NWP range from 500-2500 km.
      initial_variance: The average (over EarthSurface) variance of the random
        field. If None, this GRF always returns a zeros field and no RNGS will
        be drawn
      variance_bound: If provided, an upper bound on tuned variance values.
      tune_variance: Whether variance should be a tunable hk.parameter, or fixed
      clip: number of standard deviations at which to clip randomness to ensure
        numerical stability.
      name: Something no one cares about and we just use None.
    """
    # You must call hk.Module.__init__ before initializing this class.
    hk.Module.__init__(self, name=name)

    correlation_time_raw = hk.get_parameter(
        'correlation_time_raw', shape=(), init=hk.initializers.Constant(0.0)
    )
    correlation_length_raw = hk.get_parameter(
        'correlation_length_raw', shape=(), init=hk.initializers.Constant(0.0)
    )

    if tune_variance:
      variance_raw = hk.get_parameter(
          'variance_raw', shape=(), init=hk.initializers.Constant(0.0)
      )
    else:
      variance_raw = 0.0

    initial_variance = maybe_nondimensionalize(initial_variance, physics_specs)
    _assert_positive_or_none(initial_variance, 'initial_variance')

    if initial_variance is None:
      variance = None
    elif variance_bound in {None, 'None'}:  # Allow strings for gin.
      variance = convert_hk_param_to_positive_scalar(
          variance_raw, initial_variance
      )
    else:
      variance_bound = maybe_nondimensionalize(variance_bound, physics_specs)
      _assert_positive_or_none(variance_bound, 'variance_bound')
      _assert_positive_or_none(
          variance_bound - initial_variance, 'variance_bound - initial_variance'
      )
      variance = convert_hk_param_to_bounded_scalar(
          variance_raw,
          initial_variance,
          low=0.0,
          high=variance_bound,
      )

    # We call GaussianRandomFieldModule.__init__ rather than super().__init__
    # since we don't want to call hk.Module.__init__ twice... although doing
    # that didn't hurt anything.
    GaussianRandomField.__init__(
        self,
        coords=coords,
        dt=dt,
        physics_specs=physics_specs,
        aux_features=aux_features,
        correlation_time=convert_hk_param_to_positive_scalar(
            correlation_time_raw,
            maybe_nondimensionalize(initial_correlation_time, physics_specs),
        ),
        correlation_length=convert_hk_param_to_positive_scalar(
            correlation_length_raw,
            maybe_nondimensionalize(initial_correlation_length, physics_specs),
        ),
        variance=variance,
        clip=clip,
    )


################################################################################
# Single random fields that are derived from "stand on their own" fields.
################################################################################


@gin.register
class CenteredLognormalRandomField(GaussianRandomField):
  """A lognormal random field shifted to have mean zero."""

  @property
  def preferred_representation(self) -> PreferredRepresentation | None:
    return PreferredRepresentation.NODAL

  def _integrated_grf_variance(self) -> jax.Array | None:
    """Integrated variance of the associated GRF (not this Lognormal field)."""
    if self.variance is None:
      return None
    # If Z ~ Normal(μ, σ²), then X ~ exp(Z) has
    #  variance = (exp(σ²) - 1) exp(2μ + σ²).
    # We have centered this field, which involved setting μ = -σ² / 2.
    # => variance = exp(σ²) - 1,
    # and thus
    #  σ² = log(1 + variance)
    return jnp.log1p(self.variance) * self._surf_area

  def to_nodal_values(self, core_state: CoreRandomState) -> jax.Array:
    """Returns the ready-for-use Lognormal random field."""
    if self.variance is None:
      grf_variance = 0.0
    else:
      grf_variance = self._integrated_grf_variance() / self._surf_area
    # If Z ~ Normal(μ, σ²), then X ~ exp(Z) has mean exp(μ + σ²/2).
    # To ensure E[X] = 1, we must set μ = -σ²/2.
    x = self.coords.horizontal.to_nodal(core_state)  # ~ Normal(0, σ²)
    return jnp.expm1(x - grf_variance / 2)  # ~ Exp(Normal(-σ²/2, σ²)) - 1

  def to_modal_values(self, core_state: CoreRandomState) -> jax.Array:
    """Returns the ready-for-use Lognormal random field."""
    return self.coords.horizontal.to_modal(self.to_nodal_values(core_state))


@gin.register
class CenteredLognormalRandomFieldModule(
    CenteredLognormalRandomField, GaussianRandomFieldModule
):
  """A lognormal random hk.Module field shifted to have mean zero."""


################################################################################
# Fields made from many different fields.
################################################################################


@gin.register
class BatchGaussianRandomFieldModule(hk.Module):
  """Batch of independent GaussianRandomFieldModules.

  These GRFs are meant to be fed into a neural network as generic "signals".

  The state arrays have leading batch dim indexing independent GRFs.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      initial_correlation_times: Sequence[Quantity | str] = gin.REQUIRED,
      initial_correlation_lengths: Sequence[Quantity | str] = gin.REQUIRED,
      variances: Sequence[Quantity | str] = gin.REQUIRED,
      field_subset: Optional[Sequence[int]] = None,
      n_fixed_fields: Optional[int] = None,
      clip: float = 6.0,
      name: Optional[str] = None,
  ):
    """Constructs a BatchGaussianRandomFieldModule.

    Correlation scales are initialized to `initial_*` args and will be tuned
    by Haiku optimizers. Variance will be fixed.

    Args:
      coords: horizontal and vertical grid data.
      dt: nondimensionalized model time step.
      physics_specs: physical constants and definition of custom units.
      aux_features: additional static data.
      initial_correlation_times: timescales with units over which autoregressive
        process decorrelates. Typical values in NWP range from hours to days.
      initial_correlation_lengths: lengthscale with units over which random
        field is correlated. Typical values in NWP range from 500-2500 km.
      variances: The average (over EarthSurface) variance of the random field.
        These are fixed arrays (not tunable hk.parameters).
      field_subset: Optional nonempty subset of indices into initial parameters.
        Specifies which fields to construct. If None, use all fields. E.g.,
        field_subset=[0, 5] means form 3 GRFs from the 0th and 5th parameter
        values.
      n_fixed_fields: Number of fields that use fixed parameters. These will
        be fixed at the trailing `n_fixed_fields` initial correlations. The
        total number of fields is unchanged, since these fixed fields replace
        learnable fields.
      clip: number of standard deviations at which to clip randomness to ensure
        numerical stability.
      name: Name to show in xprof.
    """
    ## You must call hk.Module.__init__ before initializing this class.
    hk.Module.__init__(self, name=name)

    lengths = [
        len(initial_correlation_times),
        len(initial_correlation_lengths),
        len(variances),
    ]
    if len(set(lengths)) != 1:
      raise ValueError(f'Argument lengths differed: {lengths=}')
    n_fixed_fields = n_fixed_fields or 0

    # Get subset of args using `field_subset`
    if field_subset is not None:
      if not field_subset:
        raise ValueError(
            '`field_subset` must be `None` or non-empty sequence. Found'
            f' {field_subset=}'
        )
      get_subset = lambda seq: [seq[i] for i in field_subset]
      initial_correlation_lengths = get_subset(initial_correlation_lengths)
      initial_correlation_times = get_subset(initial_correlation_times)
      variances = get_subset(variances)

    logging.info(
        '[NGCM] Initializing BatchGaussianRandomFieldModule with'
        f' {initial_correlation_times=}, and {initial_correlation_lengths=},'
        f' and {variances=}'
    )

    # Get Haiku parameters.
    self._n_fields = len(variances)
    self._variances = jnp.array(
        [nondimensionalize(v, physics_specs) for v in variances]
    )

    initial_correlation_lengths = jnp.array([
        nondimensionalize(l, physics_specs) for l in initial_correlation_lengths
    ])
    correlation_lengths_raw = hk.get_parameter(
        'correlation_lengths_raw',
        shape=(self.n_fields - n_fixed_fields,),
        init=hk.initializers.Constant(0.0),
    )
    if n_fixed_fields:
      correlation_lengths_raw = jnp.concatenate([
          correlation_lengths_raw, jnp.zeros([n_fixed_fields])])
    self._correlation_lengths = convert_hk_param_to_positive_scalar(
        correlation_lengths_raw, initial_correlation_lengths
    )

    initial_correlation_times = jnp.array(
        [nondimensionalize(t, physics_specs) for t in initial_correlation_times]
    )
    correlation_times_raw = hk.get_parameter(
        'correlation_times_raw',
        shape=(self.n_fields - n_fixed_fields,),
        init=hk.initializers.Constant(0.0),
    )
    if n_fixed_fields:
      correlation_times_raw = jnp.concatenate([
          correlation_times_raw, jnp.zeros([n_fixed_fields])])
    self._correlation_times = convert_hk_param_to_positive_scalar(
        correlation_times_raw, initial_correlation_times
    )

    def make_rf(correlation_time, correlation_length, variance):
      return GaussianRandomField(
          coords=coords,
          dt=dt,
          physics_specs=physics_specs,
          aux_features=aux_features,
          correlation_time=correlation_time,
          correlation_length=correlation_length,
          variance=variance,
          clip=clip,
      )

    self._make_rf = make_rf

  @property
  def n_fields(self) -> int:
    return self._n_fields

  def unconditional_sample(self, rng: typing.PRNGKeyArray) -> RandomnessState:
    """Sample the batch GRFs unconditionally."""
    logging.info(
        '[NGCM] Calling BatchGaussianRandomFieldModule.unconditional_sample'
    )

    def _unconditional_sample_one_rf(
        key, correlation_time, correlation_length, variance
    ):
      rf = self._make_rf(correlation_time, correlation_length, variance)
      return rf.unconditional_sample(key)

    rngs = jax.random.split(rng, self.n_fields + 1)
    rngs, next_rng = rngs[:-1], rngs[-1]
    sample = jax.vmap(_unconditional_sample_one_rf)(
        rngs,
        self._correlation_times,
        self._correlation_lengths,
        self._variances,
    )
    # We have RNG keys and steps associated with each field from vmap, but
    # RandomnessState should only have a single (scalar) RNG key/step.
    return dataclasses.replace(sample, prng_key=next_rng, prng_step=0)

  def advance(self, state: RandomnessState) -> RandomnessState:
    """Updates the state of the batch of GRFs."""
    logging.info('[NGCM] Calling BatchGaussianRandomFieldModule.advance')

    def _advance_one_rf(state, correlation_time, correlation_length, variance):
      rf = self._make_rf(correlation_time, correlation_length, variance)
      return rf.advance(state)

    rng = _prng_key_for_current_advance_step(state)
    rngs = jax.random.split(rng, self.n_fields)
    steps = jnp.ones(self.n_fields, int) * state.prng_step
    advanced = jax.vmap(_advance_one_rf)(
        dataclasses.replace(state, prng_key=rngs, prng_step=steps),
        self._correlation_times,
        self._correlation_lengths,
        self._variances,
    )
    return dataclasses.replace(
        advanced, prng_key=state.prng_key, prng_step=state.prng_step + 1
    )


@gin.register
class DictOfGaussianRandomFieldModules(hk.Module):
  """Dictionary of independent GaussianRandomFieldModules.

  These GRFs are meant to be fed into a neural network as generic "signals".
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      initial_correlation_times: Sequence[Quantity | str] = gin.REQUIRED,
      initial_correlation_lengths: Sequence[Quantity | str] = gin.REQUIRED,
      variances: Sequence[Quantity | str] = gin.REQUIRED,
      field_names: Optional[Sequence[str]] = None,
      field_subset: Optional[Sequence[int]] = None,
      clip: float = 6.0,
      name: Optional[str] = None,
  ):
    """Constructs a DictOfGaussianRandomFieldModules.

    Correlation scales are initialized to `initial_*` args and will be tuned
    by Haiku optimizers. Variance will be fixed.

    Args:
      coords: horizontal and vertical grid data.
      dt: nondimensionalized model time step.
      physics_specs: physical constants and definition of custom units.
      aux_features: additional static data.
      initial_correlation_times: timescales with units over which autoregressive
        process decorrelates. Typical values in NWP range from hours to days.
      initial_correlation_lengths: lengthscale with units over which random
        field is correlated. Typical values in NWP range from 500-2500 km.
      variances: The average (over EarthSurface) variance of the random field.
        These are fixed arrays (not tunable hk.parameters).
      field_names: Optional names to give the fields. If None, the fields are
        named like "GRF0", "GRF1",...
      field_subset: Optional nonempty subset of indices into initial parameters.
        Specifies which fields to construct. If None, use all fields. E.g.,
        field_subset=[0, 5] means form 3 GRFs from the 0th and 5th parameter
        values.
      clip: number of standard deviations at which to clip randomness to ensure
        numerical stability.
      name: Name to show in xprof.
    """
    ## You must call hk.Module.__init__ before initializing this class.
    hk.Module.__init__(self, name=name)
    logging.info(
        '[NGCM] Initializing DictOfGaussianRandomFieldModules with'
        f' {initial_correlation_times=}, and {initial_correlation_lengths=},'
        f' and {variances=}'
    )

    field_names = field_names or [
        f'GRF{i}' for i in range(len(initial_correlation_times))
    ]

    lengths = [
        len(initial_correlation_times),
        len(initial_correlation_lengths),
        len(variances),
        len(field_names),
    ]
    if len(set(lengths)) != 1:
      raise ValueError(f'Argument lengths differed: {lengths=}')

    if field_subset is not None:
      if not field_subset:
        raise ValueError(
            '`field_subset` must be `None` or non-empty sequence. Found'
            f' {field_subset=}'
        )
      subset = lambda seq: [seq[i] for i in field_subset]
      field_names = subset(field_names)
      initial_correlation_lengths = subset(initial_correlation_lengths)
      initial_correlation_times = subset(initial_correlation_times)
      variances = subset(variances)

    self._field_names = tuple(field_names)

    self._random_fields = {}
    for tau, lam, var, field_name in zip(
        initial_correlation_times,
        initial_correlation_lengths,
        variances,
        self.field_names,
        strict=True,
    ):
      self._random_fields[field_name] = GaussianRandomFieldModule(
          coords,
          dt,
          physics_specs,
          aux_features,
          initial_correlation_time=tau,
          initial_correlation_length=lam,
          initial_variance=var,
          tune_variance=False,
          variance_bound=None,
          clip=clip,
          name=field_name,
      )

  @property
  def n_fields(self) -> int:
    return len(self._random_fields)

  @property
  def field_names(self) -> tuple[str, ...]:
    return self._field_names

  def unconditional_sample(self, rng: typing.PRNGKeyArray) -> RandomnessState:
    """Sample the random field unconditionally."""
    core = {}
    nodal_values = {}
    modal_values = {}
    *rngs, next_rng = jax.random.split(rng, self.n_fields + 1)
    for (name, rf), sample_key in zip(self._random_fields.items(), rngs):
      rvs = rf.unconditional_sample(sample_key)
      core[name] = rvs.core
      nodal_values[name] = rvs.nodal_value
      modal_values[name] = rvs.modal_value
    return RandomnessState(
        core=core,
        nodal_value=nodal_values,
        modal_value=modal_values,
        prng_key=next_rng,
        prng_step=0,
    )

  def advance(self, state: RandomnessState) -> RandomnessState:
    """Updates the core state of a random field."""
    core = {}
    nodal_values = {}
    modal_values = {}
    rng = _prng_key_for_current_advance_step(state)
    rngs = jax.random.split(rng, self.n_fields)
    for (name, rf), sample_key in zip(self._random_fields.items(), rngs):
      # rvs is a RandomnessState.
      rvs = rf.advance(
          RandomnessState(state.core[name], prng_key=sample_key, prng_step=0)
      )
      core[name] = rvs.core
      nodal_values[name] = rvs.nodal_value
      modal_values[name] = rvs.modal_value
    return RandomnessState(
        core=core,
        nodal_value=nodal_values,
        modal_value=modal_values,
        prng_key=state.prng_key,
        prng_step=state.prng_step + 1,
    )


class SumOfRandomFields(RandomField):
  """RandomField that is the sum of multiple fields."""

  def __init__(self, random_fields: Sequence[RandomField]):
    self._random_fields = list(random_fields)  # Shallow copy
    coords = self._random_fields[0].coords
    if any(rf.coords != coords for rf in self._random_fields):
      raise ValueError(f'All fields must have the same coords. Found {coords=}')
    super().__init__(coords)

  @property
  def preferred_representation(self) -> PreferredRepresentation | None:
    n_nodal = sum(
        rf.preferred_representation == PreferredRepresentation.NODAL
        for rf in self._random_fields
    )
    n_modal = sum(
        rf.preferred_representation == PreferredRepresentation.MODAL
        for rf in self._random_fields
    )
    if n_nodal > n_modal:
      return PreferredRepresentation.NODAL
    elif n_nodal < n_modal:
      return PreferredRepresentation.MODAL
    return None

  def unconditional_sample(self, rng: typing.PRNGKeyArray) -> RandomnessState:
    """Sample the random field unconditionally."""
    rvs = []
    *rngs, next_rng = jax.random.split(rng, len(self._random_fields) + 1)
    for rf, sample_key in zip(self._random_fields, rngs, strict=True):
      rvs.append(rf.unconditional_sample(sample_key).core)
    return RandomnessState(
        core=rvs,
        nodal_value=self.to_nodal_values(rvs),
        modal_value=self.to_modal_values(rvs),
        prng_key=next_rng,
        prng_step=0,
    )

  def advance(self, state: RandomnessState) -> RandomnessState:
    """Updates the core state of a random field."""
    rvs = []
    rng = _prng_key_for_current_advance_step(state)
    rngs = jax.random.split(rng, len(self._random_fields))
    for rf, s, k in zip(
        self._random_fields, state.core, rngs, strict=True
    ):
      rs = RandomnessState(s, prng_key=k, prng_step=state.prng_step)
      rvs.append(rf.advance(rs).core)
    return RandomnessState(
        core=rvs,
        nodal_value=self.to_nodal_values(rvs),
        modal_value=self.to_modal_values(rvs),
        prng_key=state.prng_key,
        prng_step=state.prng_step + 1,
    )

  def to_modal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Finishes `core_state` by summing components."""
    modal_sum = 0.0
    nodal_sum = 0.0
    for rf, s in zip(self._random_fields, core_state, strict=True):
      if rf.preferred_representation == PreferredRepresentation.NODAL:
        nodal_sum += rf.to_nodal_values(s)
      elif rf.preferred_representation in [PreferredRepresentation.MODAL, None]:
        modal_sum += rf.to_modal_values(s)
    return modal_sum + self.coords.horizontal.to_modal(nodal_sum)

  def to_nodal_values(self, core_state: CoreRandomState) -> typing.Array | None:
    """Finishes `core_state` by summing components."""
    modal_sum = 0.0
    nodal_sum = 0.0
    for rf, s in zip(self._random_fields, core_state, strict=True):
      if rf.preferred_representation == PreferredRepresentation.MODAL:
        modal_sum += rf.to_modal_values(s)
      elif rf.preferred_representation in [PreferredRepresentation.NODAL, None]:
        nodal_sum += rf.to_nodal_values(s)
    return nodal_sum + self.coords.horizontal.to_nodal(modal_sum)


class SumOfGaussianLikeRandomFields(SumOfRandomFields, abc.ABC):
  """Base class for sum of independent Gaussian-like random fields."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      correlation_times: Sequence[
          Union[jax.Array, Quantity, str]
      ] = gin.REQUIRED,
      correlation_lengths: Sequence[
          Union[jax.Array, Quantity, str]
      ] = gin.REQUIRED,
      variances: Sequence[Union[jax.Array, Quantity, str]] = gin.REQUIRED,
      clip: float = 6.0,
  ):
    """Constructs a SumOfGaussianLikeRandomFields."""
    n_fields = len(correlation_times)
    variances = variances or [None] * n_fields
    random_fields = []
    logging.info(
        '[NGCM] Initializing SumOfGaussianLikeRandomFields with '
        f'{variances=}, {correlation_times=}, {correlation_lengths=}'
    )
    for tau, lam, var in zip(
        correlation_times, correlation_lengths, variances, strict=True
    ):
      random_fields.append(
          self.get_cls_constructor()(
              coords,
              dt,
              physics_specs,
              aux_features,
              correlation_time=tau,
              correlation_length=lam,
              variance=var,
              clip=clip,
          )
      )

    super().__init__(random_fields)

  @abc.abstractmethod
  def get_cls_constructor(self) -> type[GaussianRandomField]:
    """Gets class constructor that is initialized with Gaussian-like kwargs."""


class SumOfGaussianLikeRandomFieldsModule(
    SumOfRandomFields, hk.Module, abc.ABC
):
  """Base class for sums of independent Gaussian-like RandomFieldModules."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      initial_correlation_times: Sequence[Quantity | str] = gin.REQUIRED,
      initial_correlation_lengths: Sequence[Quantity | str] = gin.REQUIRED,
      initial_variances: Optional[Sequence[Quantity | str]] = gin.REQUIRED,
      variance_bounds: Optional[Sequence[Quantity | str]] = gin.REQUIRED,
      clip: float = 6.0,
      name: Optional[str] = None,
  ):
    """Constructs a SumOfGaussianLikeRandomFieldsModule."""
    # You must call hk.Module.__init__ before initializing this class.
    hk.Module.__init__(self, name=name)

    n_fields = len(initial_correlation_times)
    initial_variances = initial_variances or [None] * n_fields
    variance_bounds = variance_bounds or [None] * n_fields
    random_fields = []
    for tau, lam, var, bound in zip(
        initial_correlation_times,
        initial_correlation_lengths,
        initial_variances,
        variance_bounds,
        strict=True,
    ):
      random_fields.append(
          self.get_cls_constructor()(
              coords,
              dt,
              physics_specs,
              aux_features,
              initial_correlation_time=tau,
              initial_correlation_length=lam,
              initial_variance=var,
              variance_bound=bound,
              clip=clip,
              name=name,
          )
      )
    # We call SumOfRandomFields.__init__ rather than super().__init__
    # since we don't want to call hk.Module.__init__ twice... although doing
    # that didn't hurt anything.
    SumOfRandomFields.__init__(self, random_fields)

  @abc.abstractmethod
  def get_cls_constructor(self) -> type[GaussianRandomFieldModule]:
    """Gets class constructor that is initialized with Gaussian-like kwargs."""


@gin.register
class SumOfGaussianRandomFields(SumOfGaussianLikeRandomFields):

  def get_cls_constructor(self) -> type[GaussianRandomField]:
    return GaussianRandomField


@gin.register
class SumOfGaussianRandomFieldsModule(SumOfGaussianLikeRandomFieldsModule):
  """A sum of independent GaussianRandomFieldModules."""

  def get_cls_constructor(self) -> type[GaussianRandomFieldModule]:
    return GaussianRandomFieldModule


@gin.register
class SumOfCenteredLognormalRandomFields(SumOfGaussianLikeRandomFields):

  def get_cls_constructor(self) -> type[CenteredLognormalRandomField]:
    return CenteredLognormalRandomField


@gin.register
class SumOfCenteredLognormalRandomFieldsModule(
    SumOfGaussianLikeRandomFieldsModule
):
  """A sum of independent CenteredLognormalRandomFieldModules."""

  def get_cls_constructor(self) -> type[CenteredLognormalRandomFieldModule]:
    return CenteredLognormalRandomFieldModule


################################################################################
# Helper functions for creating fields.
################################################################################


def convert_hk_param_to_positive_scalar(
    param: jax.Array,
    initial_value: Numeric,
) -> jax.Array:
  """Converts Haiku [batch] scalar parameter to scalar value using Softplus."""
  return initial_value * make_positive_scalar(param)


def convert_hk_param_to_bounded_scalar(
    param: jax.Array,
    initial_value: Numeric,
    low: Numeric,
    high: Numeric,
) -> jax.Array:
  """Converts a Haiku [batch] scalar parameter to scalar value using Sigmoid."""
  bijector = tfb.Sigmoid(low=low, high=high)
  # Since param initializes at 0, at initialization,
  #  bijector.forward(offset + param)
  #  = bijector.forward(offset)
  #  = bijector.forward(bijector.inverse(initial_value))
  #  = initial_value.
  offset = bijector.inverse(initial_value)
  return bijector.forward(offset + param)


def nondimensionalize(
    x: Union[typing.Numeric, Quantity, str],
    physics_specs: Any,
) -> typing.Numeric:
  if isinstance(x, (Quantity, str)):
    return physics_specs.nondimensionalize(Quantity(x))
  else:
    return x


def maybe_nondimensionalize(
    x: Optional[Union[typing.Numeric, Quantity, str]],
    physics_specs: Any,
) -> None | typing.Numeric:
  """Calls nondimensionalize on Quantity or str, otherwise passthrough."""
  if x == 'None':  # Allow strings for gin
    return None
  return nondimensionalize(x, physics_specs)


def _assert_positive_or_none(x: typing.Numeric | None, name: str) -> None:
  if x is None:
    return
  if x <= 0:
    raise ValueError(f'{name}={x} but should have been positive or None')



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl.testing import absltest
import jax
import jax.numpy as jnp
from neuralgcm.experimental import typing
import numpy as np


class TimedeltaTest(absltest.TestCase):

  def test_defaults(self):
    actual = typing.Timedelta()
    expected = typing.Timedelta(0, 0)
    self.assertEqual(actual, expected)

  def test_normalization(self):
    actual = typing.Timedelta(0, 24 * 60 * 60)
    expected = typing.Timedelta(1, 0)
    self.assertEqual(actual, expected)

    actual = typing.Timedelta(0, -1)
    expected = typing.Timedelta(-1, 86399)
    self.assertEqual(actual, expected)

  def test_from_timedelta64(self):
    expected = typing.Timedelta(365, 0)
    actual = typing.Timedelta.from_timedelta64(np.timedelta64(365, 'D'))
    self.assertEqual(actual, expected)

  def test_to_timedelta64(self):
    delta = typing.Timedelta(365, 0)
    expected = np.timedelta64(365, 'D')
    actual = delta.to_timedelta64()
    self.assertEqual(actual, expected)

  def test_addition(self):
    delta = typing.Timedelta(1, 12 * 60 * 60)
    actual = delta + delta
    expected = typing.Timedelta(3, 0)
    self.assertEqual(actual, expected)

    with self.assertRaises(TypeError):
      expected + 60

  def test_negation(self):
    delta = typing.Timedelta(1, 6 * 60 * 60)
    actual = -delta
    expected = typing.Timedelta(-2, 18 * 60 * 60)
    self.assertEqual(actual, expected)

  def test_subtraction(self):
    delta = typing.Timedelta(1, 12 * 60 * 60)
    actual = delta - delta
    expected = typing.Timedelta(0, 0)
    self.assertEqual(actual, expected)

  def test_multiplication(self):
    delta = typing.Timedelta(1, 12 * 60 * 60)
    expected = typing.Timedelta(3, 0)

    actual = delta * 2
    self.assertEqual(actual, expected)

    actual = 2 * delta
    self.assertEqual(actual, expected)

    with self.assertRaises(TypeError):
      delta * delta

  def test_pytree(self):
    # pytree transformations do not normalize seconds
    delta = typing.Timedelta(1, 12 * 60 * 60)
    expected = typing.Timedelta(2)
    expected.seconds = 24 * 60 * 60  # cannot construct directly
    actual = jax.tree.map(lambda x: 2 * x, delta)
    self.assertEqual(actual, expected)

  def test_vmap(self):
    delta = typing.Timedelta(days=jnp.arange(2), seconds=jnp.arange(2))
    result = jax.vmap(lambda x: x)(delta)
    self.assertIsInstance(result, typing.Timedelta)
    np.testing.assert_array_equal(result.days, delta.days)
    np.testing.assert_array_equal(result.seconds, delta.seconds)


class TimestampTest(absltest.TestCase):

  def test_from_datetime64(self):
    expected = typing.Timestamp(typing.Timedelta(365, 0))
    actual = typing.Timestamp.from_datetime64(np.datetime64('1971-01-01'))
    self.assertEqual(expected, actual)

  def test_to_datetime64(self):
    timestamp = typing.Timestamp(typing.Timedelta(365, 0))
    expected = np.datetime64('1971-01-01')
    actual = timestamp.to_datetime64()
    self.assertEqual(expected, actual)

  def test_datetime64_through_int32_roundtrip(self):
    original = np.datetime64('2024-01-01T00:00:00', 'ns')
    timestamp = typing.Timestamp.from_datetime64(original)
    timestamp_int32 = jax.tree.map(np.int32, timestamp)
    restored = timestamp_int32.to_datetime64()
    self.assertEqual(original, restored)

  def test_addition(self):
    delta = typing.Timedelta(1, 0)
    timestamp = typing.Timestamp(delta)
    expected = typing.Timestamp(delta * 2)

    actual = timestamp + delta
    self.assertEqual(actual, expected)

    actual = delta + timestamp
    self.assertEqual(actual, expected)

    with self.assertRaises(TypeError):
      timestamp + 1

    with self.assertRaises(TypeError):
      timestamp + timestamp

  def test_subtraction(self):
    first = typing.Timestamp.from_datetime64(
        np.datetime64('2024-01-01T00:00:00')
    )
    second = typing.Timestamp.from_datetime64(
        np.datetime64('2023-01-01T00:00:00')
    )
    delta = typing.Timedelta(365, 0)
    actual = first - delta
    self.assertEqual(actual, second)
    actual = -delta + first
    self.assertEqual(actual, second)
    actual = first - second
    self.assertEqual(actual, delta)
    actual = second - first
    self.assertEqual(actual, -delta)

  def test_vmap(self):
    stamp = typing.Timestamp(
        typing.Timedelta(days=jnp.arange(2), seconds=jnp.arange(2))
    )
    result = jax.vmap(lambda x: x)(stamp)
    self.assertIsInstance(result, typing.Timestamp)
    np.testing.assert_array_equal(result.delta.days, stamp.delta.days)
    np.testing.assert_array_equal(result.delta.seconds, stamp.delta.seconds)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines `forcing` modules that produce time-dependent focing values."""

from __future__ import annotations

import functools
import logging
from typing import Any, Optional, Union

from dinosaur import coordinate_systems
from dinosaur import scales
from dinosaur import typing
import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import transforms
import numpy as np

tree_leaves = jax.tree_util.tree_leaves
tree_map = jax.tree_util.tree_map
units = scales.units

Pytree = typing.Pytree
ForcingData = typing.ForcingData
ForcingFn = typing.ForcingFn
Forcing = typing.Forcing
TransformModule = typing.TransformModule
Quantity = units.Quantity
QuantityOrStr = Union[str, scales.Quantity]


# _FORCING_ERRORS global will store errors obtained during a io_callback.
# The user can periodically call _check_errors to see if errors have accumulated
# TODO(langmore) Use a more universal mechanism (not just in forcings.py) to
# handle errors, if we like this, then make public.
_FORCING_ERRORS = []

# pylint: disable=logging-fstring-interpolation


class ForcingDataError(Exception):
  """To raise when an error is encountered with forcing data."""


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class NoForcing(hk.Module):
  """Module that returns an empty Forcing object."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      time_axis: int = 0,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    del coords, dt, physics_specs, aux_features, time_axis

  def __call__(
      self,
      forcing_data: ForcingData,
      sim_time: float,
  ) -> Forcing:
    """Returns forcings at the specified sim_time."""
    del forcing_data, sim_time
    return {}


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class DynamicDataForcing(hk.Module):
  """Modules that returns forcing values by querying time-varying data.

  Input to __call__ `sim_time` must match a value in forcing_data['sim_time']
  within dt_tolerance, or else it returns nan for all pytree values.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      inputs_to_units_mapping: dict[str, str],
      forcing_transform: TransformModule = transforms.IdentityTransform,
      time_axis: int = 0,
      data_time_step: float | QuantityOrStr | None = None,
      dt_tolerance: Union[float, QuantityOrStr] = '1 hour',
      # TODO(langmore) Remove checking once bug arising from http://cl/624039690
      # is fixed.
      check_sim_time_errors: bool = False,
      name: Optional[str] = None,
  ):
    logging.info(f'[NGCM] Initializing DynamicDataForcing with {dt_tolerance=}')
    # TODO(shoyer): remove data_time_step entirely, once we're sure that no
    # saved checkpoints that we care about will break.
    del data_time_step  # no longer used
    super().__init__(name=name)
    self.time_axis = time_axis
    self.nondim_transform_fn = transforms.NondimensionalizeTransform(
        coords,
        dt,
        physics_specs,
        aux_features,
        input_coords=None,
        inputs_to_units_mapping=inputs_to_units_mapping,
    )
    self.forcing_transform_fn = forcing_transform(
        coords, dt, physics_specs, aux_features
    )
    if isinstance(dt_tolerance, (str, scales.Quantity)):
      dt_tolerance = physics_specs.nondimensionalize(
          scales.Quantity(dt_tolerance)
      )
    self.dt_tolerance = dt_tolerance
    self._check_sim_time_errors = check_sim_time_errors

  def __call__(
      self,
      forcing_data: ForcingData,
      sim_time: float,
  ) -> Forcing:
    """Returns forcings at the specified sim_time."""
    forcing_data = self.nondim_transform_fn(forcing_data)

    times = forcing_data['sim_time']
    approx_index = jnp.interp(sim_time, times, jnp.arange(times.size))
    index = jnp.round(approx_index).astype('int32')

    # Slice leaf values by index
    field_index_fn = functools.partial(
        jax.lax.dynamic_index_in_dim,
        index=index,
        axis=self.time_axis,
        keepdims=False,
    )
    _assert_no_scalars(forcing_data)
    forcing = tree_map(field_index_fn, forcing_data)

    # Replace leaf values with nan if forcing['sim_time'] does not match
    # the requested sim_time value within dt_tolerance.
    abs_error = jnp.abs(forcing['sim_time'] - sim_time)
    is_valid = abs_error < self.dt_tolerance
    forcing = jax.tree_util.tree_map(
        lambda x: jnp.where(is_valid, x, jnp.nan), forcing
    )

    # Also add errors (if any) to _FORCING_ERRORS so _check_errors can be called
    # to raise.
    if self._check_sim_time_errors:
      jax.experimental.io_callback(
          _check_sim_time_close_to_forcing_sim_time,
          None,  # Returns None
          sim_time=sim_time,
          forcing_sim_time=forcing['sim_time'],
          tolerance=self.dt_tolerance,
      )
    return self.forcing_transform_fn(forcing)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class PersistenceDataForcing(hk.Module):
  """Modules that returns forcing using first time index of forcing_data."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      inputs_to_units_mapping: dict[str, str],
      forcing_transform: TransformModule = transforms.IdentityTransform,
      time_axis: int = 0,
      name: Optional[str] = None,
  ):
    logging.info('[NGCM] Initializing PersistenceDataForcing')
    super().__init__(name=name)
    self.time_axis = time_axis
    self.nondim_transform_fn = transforms.NondimensionalizeTransform(
        coords,
        dt,
        physics_specs,
        aux_features,
        input_coords=None,
        inputs_to_units_mapping=inputs_to_units_mapping,
    )
    self.forcing_transform_fn = forcing_transform(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      forcing_data: ForcingData,
      sim_time: float,
  ) -> Forcing:
    """Returns forcings from the first time index of sim_time."""
    del sim_time  # unused
    forcing_data = self.nondim_transform_fn(forcing_data)
    idx = 0

    # Slice leaf values by index
    field_index_fn = functools.partial(
        jax.lax.dynamic_index_in_dim,
        index=idx,
        axis=self.time_axis,
        keepdims=False,
    )
    _assert_no_scalars(forcing_data)
    forcing = tree_map(field_index_fn, forcing_data)
    return self.forcing_transform_fn(forcing)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class IncrementSSTForcingTransform(hk.Module):
  """Transform Forcing by uniformly incrementing sea surface temperature."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      temperature_change: Quantity,
      key: str = 'sea_surface_temperature',
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    del coords, dt, aux_features  # unused
    self.temperature_change = physics_specs.nondimensionalize(
        units.Quantity(temperature_change)
    )
    self.key = key

  def __call__(self, forcing: Forcing) -> Forcing:
    assert isinstance(forcing, dict)
    forcing = forcing.copy()
    forcing[self.key] = forcing[self.key] + self.temperature_change
    return forcing


def _assert_no_scalars(tree: Pytree):
  dims = tree_map(lambda x: len(jnp.shape(x)), tree)
  if not all(d > 0 for d in tree_leaves(dims)):
    raise ValueError(f'Scalar shapes encountered: {dims=}')


# TODO(langmore) Use a more universal mechanism (not just in forcings.py) to
# handle errors, if we like this, then make public.
def _check_sim_time_close_to_forcing_sim_time(
    sim_time: np.ndarray,
    forcing_sim_time: np.ndarray,
    tolerance: float,
) -> None:
  """Checks |sim_time - forcing_sim_time| < tolerance add to _FORCING_ERRORS."""
  abs_error = np.abs(forcing_sim_time - sim_time)
  if abs_error < tolerance:
    return
  err_msg = (
      f'{sim_time=} differed from {forcing_sim_time=} by {abs_error=} which is '
      f'> {tolerance=}'
  )
  _FORCING_ERRORS.append(err_msg)


# TODO(langmore) Use a more universal mechanism (not just in forcings.py) to
# handle errors, if we like this, then make public.
def _check_errors(  # pylint: disable=dangerous-default-value
    max_to_print: int = 4,
    err_list: list[str] = _FORCING_ERRORS,
) -> None:
  """Check err_list and raise ForcingDataError if nonempty."""
  n_err = len(err_list)
  if n_err:
    raise ForcingDataError(
        f'ForcingDataError found: {n_err} exceptions: '
        f'The first {min(n_err, max_to_print)} are: '
        f'{", ".join(err_list[:max_to_print])}'
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for standard_layers.py."""

from absl.testing import absltest
from absl.testing import parameterized
from flax import nnx
import jax
import jax.numpy as jnp
from neuralgcm.experimental import standard_layers
import numpy as np


class StandardLayersTest(parameterized.TestCase):
  """Tests standard layers primitives."""

  @parameterized.parameters(
      dict(
          input_size=11,
          output_size=3,
          num_hidden_units=6,
          num_hidden_layers=1,
          use_bias=True,
      ),
      dict(
          input_size=2,
          output_size=6,
          num_hidden_units=6,
          num_hidden_layers=0,
          use_bias=True,
      ),
      dict(
          input_size=8,
          output_size=5,
          num_hidden_units=13,
          num_hidden_layers=4,
          use_bias=False,
      ),
  )
  def test_mlp_uniform(
      self,
      input_size: int,
      output_size: int,
      num_hidden_units: int,
      num_hidden_layers: int,
      use_bias: bool,
  ):
    """Tests output_shape, number and initialization of params in MlpUniform."""
    mlp = standard_layers.MlpUniform(
        input_size=input_size,
        output_size=output_size,
        hidden_size=num_hidden_units,
        n_hidden_layers=num_hidden_layers,
        use_bias=use_bias,
        rngs=nnx.Rngs(0),
    )
    inputs = np.random.uniform(size=input_size)
    params = nnx.state(mlp, nnx.Param)
    out = mlp(inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size,))
    with self.subTest('total_params_count'):
      expected = count_mlp_uniform_params(
          input_size=input_size,
          num_hidden_units=num_hidden_units,
          num_hidden_layers=num_hidden_layers,
          output_size=output_size,
          use_bias=use_bias,
      )
      actual = sum([np.prod(x.shape) for x in jax.tree.leaves(params)])
      self.assertEqual(actual, expected)

    if num_hidden_layers > 1:
      with self.subTest('independent_params'):
        for previous_layer, layer in zip(mlp.layers[1:-3], mlp.layers[2:-2]):
          kernel_diff = previous_layer.kernel.value - layer.kernel.value
          self.assertGreater(jnp.linalg.norm(kernel_diff), 1e-1)

  @parameterized.parameters(
      dict(
          input_size=8,
          output_size=2,
          kernel_size=3,
      ),
      dict(
          input_size=9,
          output_size=4,
          kernel_size=5,
          use_bias=False,
      ),
      dict(
          input_size=5,
          output_size=11,
          kernel_size=3,
          use_bias=False,
      ),
      dict(
          input_size=2,
          output_size=4,
          kernel_size=1,
      ),
  )
  def test_conv_level(
      self,
      input_size: int,
      output_size: int,
      kernel_size: int,
      use_bias: bool = True,
  ):
    """Tests output_shape and number of params in ConvLevel."""
    level_size = 24
    input_shape = (input_size, level_size)
    inputs = jax.random.uniform(jax.random.PRNGKey(42), shape=input_shape)
    conv_level_layer = standard_layers.ConvLevel(
        input_size=input_size,
        output_size=output_size,
        kernel_size=kernel_size,
        use_bias=use_bias,
        rngs=nnx.Rngs(0),
    )
    params = nnx.state(conv_level_layer, nnx.Param)
    out = conv_level_layer(inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size, level_size))
    with self.subTest('total_params_count'):
      expected = count_conv_params(
          input_size=input_size,
          kernel_size=kernel_size,
          output_size=output_size,
          use_bias=use_bias,
      )
      actual = sum([np.prod(x.shape) for x in jax.tree.leaves(params)])
      self.assertEqual(actual, expected)

  @parameterized.parameters(
      dict(
          input_size=8,
          output_size=2,
          channels=(6,),
          kernel_sizes=6,
      ),
      dict(
          input_size=9,
          output_size=4,
          channels=(4, 5),
          kernel_sizes=(3, 3, 3),
          use_bias=False,
      ),
      dict(
          input_size=2,
          output_size=4,
          channels=(7, 3),
          kernel_sizes=(5, 3, 6),
          kernel_dilations=(3, 1, 1),
      ),
  )
  def test_cnn_level_shape(
      self,
      input_size: int,
      output_size: int,
      channels: tuple[int, ...],
      kernel_sizes: int | tuple[int, ...],
      kernel_dilations: int | tuple[int, ...] = 1,
      use_bias: bool = True,
  ):
    """Tests that VerticalConvNet outputs and params have expected shapes."""
    n_levels = 13
    input_shape = (input_size, n_levels)
    inputs = jax.random.uniform(jax.random.PRNGKey(42), shape=input_shape)
    cnn_level_layer = standard_layers.CnnLevel(
        input_size=input_size,
        output_size=output_size,
        channels=channels,
        kernel_sizes=kernel_sizes,
        use_bias=use_bias,
        kernel_dilations=kernel_dilations,
        rngs=nnx.Rngs(0),
    )
    params = nnx.state(cnn_level_layer, nnx.Param)
    out = cnn_level_layer(inputs)
    with self.subTest('output_shape'):
      actual_shape = out.shape
      expected_shape = (output_size, n_levels)
      self.assertEqual(actual_shape, expected_shape)

    with self.subTest('total_params_count'):
      if isinstance(kernel_sizes, int):
        kernel_sizes = (kernel_sizes,) * (len(channels) + 1)
      in_out_and_kernels = zip(
          (input_size,) + channels,
          channels + (output_size,),
          kernel_sizes,
      )
      expected_by_part = [
          count_conv_params(din, kernel_size, dout, use_bias)
          for din, dout, kernel_size in in_out_and_kernels
      ]
      expected = sum(expected_by_part)
      actual = sum([np.prod(x.shape) for x in jax.tree.leaves(params)])
      self.assertEqual(actual, expected)

  @parameterized.parameters(
      dict(
          input_size=1,
          input_lonlat_shape=(54, 54),
          output_size=1,
          kernel_size=(5, 5),
          dilation=1,
          use_bias=True,
      ),
      dict(
          input_size=2,
          input_lonlat_shape=(54, 27),
          output_size=1,
          kernel_size=(3, 3),
          dilation=4,
          use_bias=False,
      ),
      dict(
          input_size=6,
          input_lonlat_shape=(33, 87),
          output_size=5,
          kernel_size=(3, 3),
          dilation=2,
          use_bias=True,
      ),
  )
  def test_conv_lon_lat(
      self,
      input_size: int,
      input_lonlat_shape: tuple[int, int],
      output_size: int,
      kernel_size: tuple[int, int],
      dilation: int,
      use_bias: bool,
  ):
    conv_layer = standard_layers.ConvLonLat(
        input_size=input_size,
        output_size=output_size,
        kernel_size=kernel_size,
        dilation=dilation,
        use_bias=use_bias,
        rngs=nnx.Rngs(0),
    )
    params = nnx.state(conv_layer, nnx.Param)

    with self.subTest('output_shape'):
      inputs = jnp.ones((input_size,) + input_lonlat_shape)
      outputs = conv_layer(inputs)
      self.assertEqual(outputs.shape, (output_size,) + inputs.shape[1:])

    with self.subTest('total_params_count'):
      expected = count_conv_params(
          input_size, kernel_size, output_size, use_bias
      )
      actual = sum([np.prod(x.shape) for x in jax.tree.leaves(params)])
      np.testing.assert_allclose(actual, expected)
      self.assertEqual(actual, expected)

  @parameterized.parameters(
      dict(
          lon_lat_shape=(10, 10),
      ),
      dict(
          lon_lat_shape=(15, 17),
      ),
  )
  def test_conv_lon_lat_padding(self, lon_lat_shape: tuple[int, int]):
    conv = standard_layers.ConvLonLat(
        input_size=1,
        output_size=1,
        kernel_size=(3, 3),
        dilation=1,
        use_bias=False,
        rngs=nnx.Rngs(0),
    )
    lon_size, lat_size = lon_lat_shape

    test_in_wrap = jnp.broadcast_to(
        jnp.arange(lon_size, dtype=float)[:, jnp.newaxis],
        (1,) + lon_lat_shape,  # arange repeated along longitude.
    )

    test_in_reflect = jnp.broadcast_to(
        jnp.arange(lat_size, dtype=float)[jnp.newaxis, :],
        (1,) + lon_lat_shape,  # arange repeated along latitude.
    )

    with self.subTest('wrap_padding'):
      kernel_select_above = jnp.expand_dims(
          jnp.array([[0.0, 1.0, 0.0],
                     [0.0, 0.0, 0.0],
                     [0.0, 0.0, 0.0]]),
          axis=(-1, -2),
      )
      conv.conv_layer.kernel.value = kernel_select_above
      output = conv(test_in_wrap)
      np.testing.assert_allclose(output[:, 0], test_in_wrap[:, -1])
    with self.subTest('reflect_padding'):
      kernel_select_left = jnp.expand_dims(
          jnp.array([[0.0, 0.0, 0.0],
                     [1.0, 0.0, 0.0],
                     [0.0, 0.0, 0.0]]),
          axis=(-1, -2),
      )
      conv.conv_layer.kernel.value = kernel_select_left
      output = conv(test_in_reflect)
      np.testing.assert_allclose(output[:, :, 0], test_in_reflect[:, :, 1])


def count_mlp_uniform_params(
    input_size: int,
    num_hidden_units: int,
    num_hidden_layers: int,
    output_size: int,
    use_bias: bool,
):
  """Returns the number of parameters in MlpUniform based on settings."""
  n_input = input_size * num_hidden_units + num_hidden_units * use_bias
  n_hidden = num_hidden_units * num_hidden_units + num_hidden_units * use_bias
  n_output = num_hidden_units * output_size + output_size * use_bias
  return n_input + n_hidden * (num_hidden_layers - 1) + n_output


def count_conv_params(
    input_size: int, kernel_size: int, output_size: int, use_bias: bool
):
  """Returns the number of parameters in ConvLonLat based on settings."""
  n_w_params = np.prod(kernel_size) * input_size * output_size
  n_b_params = output_size
  return n_w_params + n_b_params * use_bias


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of perturbation modules."""
import abc
from typing import Any, Callable
from dinosaur import coordinate_systems
from dinosaur import pytree_utils
from dinosaur import spherical_harmonic
from dinosaur import typing
import gin
import jax
import jax.numpy as jnp

from neuralgcm import transforms

Pytree = typing.Pytree
PerturbationFn = Callable[..., Pytree]
PerturbationModule = Callable[..., PerturbationFn]


_ALLOWED_PERTURBATION_BASIS = (
    # Converts vorticity/divergence to u/v then perturbs.
    'uv',

    # Perturbs in whatever the state is in (typically vorticity/divergence).
    'generic',
)

# We ♥ λ's
# pylint: disable=g-long-lambda


@gin.register
class NoPerturbation:
  """No-op perturbation that introduces no perturbation to `inputs`."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
  ):
    """Initializes a random field."""
    del coords, dt, physics_specs, aux_features  # unused.

  def __call__(
      self,
      inputs: typing.Pytree,
      state: typing.Pytree,
      randomness: typing.Pytree,
  ) -> typing.Pytree:
    """Updates the state of a random field."""
    del state, randomness  # unused.
    return inputs


class BasePerturbation(abc.ABC):
  """Base class for perturbations."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Any,
      randomness_transform_module: transforms.TransformModule = (
          transforms.IdentityTransform
      ),
      return_modal: bool = True,
      perturbation_basis: str = 'generic',
  ):
    """Initializes module to perturb random fields.

    Args:
      coords: Model coordinate system.
      dt: Time step.
      physics_specs:
      aux_features:
      randomness_transform_module: Module that transforms jax.Array of random
        variables before converting to nodal.
      return_modal: Whether results should be returned in modal space.
      perturbation_basis: Whether to perturb wind in "uv" or "generic" basis.
    """
    self.coords = coords
    self.randomness_transform_fn = randomness_transform_module(
        coords, dt, physics_specs, aux_features
    )
    self.return_modal = return_modal
    self.to_modal = coords.horizontal.to_modal
    self.to_nodal = coords.horizontal.to_nodal
    self.maybe_to_modal = lambda tr: coordinate_systems.maybe_to_modal(
        tr, coords
    )
    self.maybe_to_nodal = lambda tr: coordinate_systems.maybe_to_nodal(
        tr, coords
    )
    if perturbation_basis not in _ALLOWED_PERTURBATION_BASIS:
      raise ValueError(
          f'{perturbation_basis=} which was not in '
          f'{_ALLOWED_PERTURBATION_BASIS=}'
      )
    self.perturbation_basis = perturbation_basis

  def __call__(
      self,
      inputs: typing.Pytree,
      state: typing.Pytree,
      randomness: typing.Pytree,
  ) -> typing.Pytree:
    """Updates the state of a random field."""
    del state  # unused.
    # TODO(dkochkov) allow pytree randomness in addition to broadcasting option.

    if self.perturbation_basis == 'generic':
      return self._perturb_in_generic_coordinates(inputs, randomness)
    elif self.perturbation_basis == 'uv':
      return self._perturb_in_uv_coordinates(inputs, randomness)

  @abc.abstractmethod
  def _perturb_core(
      self,
      inputs: typing.Pytree,
      randomness: typing.Pytree,
  ) -> typing.Pytree:
    """Perturbs inputs using randomness."""

  def _perturb_in_generic_coordinates(
      self,
      inputs: typing.Pytree,
      randomness: typing.Pytree,
  ) -> typing.Pytree:
    """Perturb `inputs` in (vorticity, divergence) coordinate system."""
    nodal_inputs = self.maybe_to_nodal(inputs)
    nodal_randomness = self.maybe_to_nodal(randomness)

    nodal_randomness = self.randomness_transform_fn(
        pytree_utils.tree_map_over_nonscalars(
            # Broadcast randomness so that self.randomness_transform_fn can use
            # the shape of x to determine what to do.
            lambda x: jnp.broadcast_to(nodal_randomness, x.shape),
            nodal_inputs,
            scalar_fn=jnp.zeros_like,
        )
    )

    perturbed_nodal_inputs = self._perturb_core(nodal_inputs, nodal_randomness)
    if self.return_modal:
      return self.to_modal(perturbed_nodal_inputs)
    else:
      return perturbed_nodal_inputs

  def _perturb_in_uv_coordinates(
      self,
      inputs: typing.Pytree,
      randomness: typing.Pytree,
  ) -> typing.Pytree:
    """Perturb `inputs` in (u, v) coordinate system."""
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)

    # Remove vorticity/divergence from inputs and replace with u/v.
    vordiv = self.maybe_to_modal({
        'vorticity': inputs.pop('vorticity'),
        'divergence': inputs.pop('divergence'),
    })
    u_nodal, v_nodal = spherical_harmonic.vor_div_to_uv_nodal(
        grid=self.coords.horizontal,
        vorticity=vordiv['vorticity'],
        divergence=vordiv['divergence'],
        clip=True,
    )
    nodal_inputs = self.maybe_to_nodal(inputs)  # Recall we popped vor/div.
    nodal_inputs['u'] = u_nodal
    nodal_inputs['v'] = v_nodal

    # Perturb in u/v space
    nodal_randomness = self.maybe_to_nodal(randomness)
    nodal_randomness = self.randomness_transform_fn(
        pytree_utils.tree_map_over_nonscalars(
            # Broadcast randomness so that self.randomness_transform_fn can use
            # the shape of x to determine what to do.
            lambda x: jnp.broadcast_to(nodal_randomness, x.shape),
            nodal_inputs,
            scalar_fn=jnp.zeros_like,
        )
    )
    perturbed_nodal_inputs = self._perturb_core(nodal_inputs, nodal_randomness)

    # Transform perturbed u/v to vor/div (modal).
    vorticity, divergence = spherical_harmonic.uv_nodal_to_vor_div_modal(
        grid=self.coords.horizontal,
        u_nodal=perturbed_nodal_inputs.pop('u'),
        v_nodal=perturbed_nodal_inputs.pop('v'),
        clip=True,
    )

    # Insert vorticity/divergence into perturbed_inputs in the right space.
    if self.return_modal:
      perturbed_inputs = self.to_modal(perturbed_nodal_inputs)
      perturbed_inputs['vorticity'] = vorticity
      perturbed_inputs['divergence'] = divergence
    else:
      perturbed_inputs = perturbed_nodal_inputs.copy()
      perturbed_inputs['vorticity'] = self.to_nodal(vorticity)
      perturbed_inputs['divergence'] = self.to_nodal(divergence)

    return from_dict_fn(perturbed_inputs)


@gin.register
class ProportionalPerturbation(BasePerturbation):
  """Perturbation that scales inputs by 1 + randomness."""

  def _perturb_core(
      self,
      inputs: typing.Pytree,
      randomness: typing.Pytree,
  ) -> typing.Pytree:
    """Multiplies inputs by (1 + randomness)."""
    return jax.tree_util.tree_map(lambda x, y: x * (1 + y), inputs, randomness)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines `encoder` modules that map input trajectories to model states.

All encoder modules return the encoder-specific model state that represents the
state of the system at the latest time provided in the input trajectory.
The inputs are expected to consist of arrays with `time` as a leading axis.
"""

# TODO(dkochkov) make all encoders take in trajectories and return ModelState.

import functools
from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union

from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import pytree_utils
from dinosaur import scales
from dinosaur import shallow_water
from dinosaur import spherical_harmonic
from dinosaur import typing
from dinosaur import vertical_interpolation
from dinosaur import weatherbench_utils
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import features
from neuralgcm import mappings
from neuralgcm import orographies
from neuralgcm import perturbations
from neuralgcm import stochastic
from neuralgcm import transforms
import numpy as np


Array = Union[np.ndarray, jnp.ndarray]
DataState = typing.DataState
FeaturesModule = features.FeaturesModule
FilterModule = Callable[..., typing.PyTreeFilterFn]
Forcing = typing.Forcing
MappingModule = mappings.MappingModule
PyTreeState = typing.PyTreeState
ModelState = typing.ModelState
TransformModule = typing.TransformModule
OrographyModule = orographies.OrographyModule
PerturbationModule = perturbations.PerturbationModule
RandomnessModule = stochastic.RandomnessModule

# We ♥ λ's
# pylint: disable=g-long-lambda


@gin.register
class EncoderIdentityTransform(hk.Module):
  """Transformation that returns inputs without modification."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    del coords, dt, physics_specs, aux_features, input_coords

  def __call__(self, inputs: PyTreeState) -> PyTreeState:
    return inputs


@gin.register
class EncoderFilterTransform(hk.Module):
  """Transformation that returns truncated and filtered modal inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      filter_modules: Sequence[FilterModule] = tuple(),
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.filter_fns = [
        module(coords, dt, physics_specs, aux_features)
        for module in filter_modules
    ]

  def __call__(self, inputs: PyTreeState) -> PyTreeState:
    for filter_fn in self.filter_fns:
      inputs = filter_fn(inputs)
    return inputs


@gin.register
class InputClipTransform(hk.Module):
  """Filter that clips highest total wavenumber the input state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      wavenumbers_to_clip: int = 1,
      name: Optional[str] = None,
  ):
    """See `time_integration.exponential_filter` for details."""
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.input_coords = input_coords
    self.wavenumbers_to_clip = wavenumbers_to_clip

  def __call__(self, inputs: PyTreeState) -> PyTreeState:
    return self.input_coords.horizontal.clip_wavenumbers(
        inputs, self.wavenumbers_to_clip
    )


@gin.register
class InputNodalToModalTransform(hk.Module):
  """Transformation that converts nodal inputs to modal representation."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.input_coords = input_coords

  def __call__(self, inputs: PyTreeState) -> PyTreeState:
    to_modal_fn = self.input_coords.horizontal.to_modal
    downsample_fn = coordinate_systems.get_spectral_downsample_fn(
        self.input_coords, self.coords, expect_same_vertical=False
    )
    return jax.tree_util.tree_map(
        lambda x: downsample_fn(to_modal_fn(x)), inputs
    )


@gin.register
class ModalInputLearnedAdaptorTransform(hk.Module):
  """Transformation using a tower to adapt modal inputs to the model domain."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      modal_to_nodal_features_module: FeaturesModule,
      nodal_mapping_module: MappingModule,
      output_transform_module: TransformModule,
      name: Optional[str] = None,
  ):
    del input_coords  # unused.
    super().__init__(name=name)
    self.coords = coords
    self.modal_to_nodal_features_fn = modal_to_nodal_features_module(
        coords, dt, physics_specs, aux_features
    )
    self.nodal_mapping_module = nodal_mapping_module
    self.output_transform_fn = output_transform_module(
        coords, dt, physics_specs, aux_features
    )
    self.get_nodal_shape_fn = lambda x: coordinate_systems.get_nodal_shapes(
        x, coords
    )

  def __call__(self, inputs: PyTreeState) -> PyTreeState:
    """Applies transform to modal inputs, returns modal outputs."""
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    prediction_shapes = jax.tree_util.tree_map(self.get_nodal_shape_fn, inputs)
    # if `inputs` contain `sim_time` - remove it from corrections.
    sim_time_shape = prediction_shapes.pop('sim_time', None)
    net = self.nodal_mapping_module(prediction_shapes)
    nodal_input_features = self.modal_to_nodal_features_fn(inputs, None)
    nodal_corrections = self.output_transform_fn(net(nodal_input_features))
    corrections = self.coords.horizontal.to_modal(nodal_corrections)
    if sim_time_shape is not None:
      corrections['sim_time'] = 0.0
    outputs = jax.tree_util.tree_map(lambda x, y: x + y, inputs, corrections)
    return from_dict_fn(outputs)


@gin.register
class NodalInputLearnedAdaptorTransform(hk.Module):
  """Transformation using a tower to adapt nodal inputs to the model domain."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      nodal_to_nodal_features_module: FeaturesModule,
      nodal_mapping_module: MappingModule,
      output_transform_module: TransformModule,
      name: Optional[str] = None,
  ):
    del input_coords  # unused.
    super().__init__(name=name)
    self.coords = coords
    self.nodal_to_nodal_features_fn = nodal_to_nodal_features_module(
        coords, dt, physics_specs, aux_features
    )
    self.nodal_mapping_module = nodal_mapping_module
    self.output_transform_fn = output_transform_module(
        coords, dt, physics_specs, aux_features
    )
    self.get_nodal_shape_fn = lambda x: coordinate_systems.get_nodal_shapes(
        x, coords
    )

  def __call__(self, inputs: PyTreeState) -> PyTreeState:
    """Applies transform to nodal inputs, returns nodal outputs."""
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    prediction_shapes = jax.tree_util.tree_map(self.get_nodal_shape_fn, inputs)
    # if `inputs` contain `sim_time` - remove it from corrections.
    sim_time_shape = prediction_shapes.pop('sim_time', None)
    net = self.nodal_mapping_module(prediction_shapes)
    input_features = self.nodal_to_nodal_features_fn(inputs, None)
    corrections = self.output_transform_fn(net(input_features))
    if sim_time_shape is not None:
      corrections['sim_time'] = 0.0
    outputs = jax.tree_util.tree_map(lambda x, y: x + y, inputs, corrections)
    return from_dict_fn(outputs)


@gin.register
class EncoderCombinedTransform(hk.Module):
  """Module that applies multiple transformations sequentially."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      input_coords: coordinate_systems.CoordinateSystem,
      transforms: Tuple[TransformModule, ...] = tuple(),  # pylint: disable=redefined-outer-name
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.transform_fns = [
        module(coords, dt, physics_specs, aux_features, input_coords)
        for module in transforms
    ]

  def __call__(self, inputs: PyTreeState) -> PyTreeState:
    for transform_fn in self.transform_fns:
      inputs = transform_fn(inputs)
    return inputs


@gin.register
class ShallowWaterStateEncoder(hk.Module):
  """Encoder that extracts shallow_water.State pair from inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      time_axis: int = 0,
      transform_module: TransformModule = EncoderIdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.slice_fn = functools.partial(
        pytree_utils.slice_along_axis, axis=time_axis, idx=-1)
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features, input_coords
    )

  def __call__(
      self, inputs: DataState, forcing: Forcing
  ) -> shallow_water.State:
    del forcing
    state = self.transform_fn(shallow_water.State(**self.slice_fn(inputs)))
    return ModelState(state)


@gin.register
class ShallowWaterLeapfrogEncoder(hk.Module):
  """Encoder that extracts shallow_water.State pair from inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      time_axis: int = 0,
      transform_module: TransformModule = EncoderIdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.slice_fn = functools.partial(
        pytree_utils.slice_along_axis, axis=time_axis, idx=slice(-2, None))
    self.time_axis = time_axis
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features, input_coords
    )

  def __call__(
      self, inputs: DataState, forcing: Forcing
  ) -> Tuple[shallow_water.State, ...]:
    del forcing
    last_two_frames = pytree_utils.split_axis(
        self.slice_fn(inputs), self.time_axis
    )
    state = self.transform_fn(
        tuple(shallow_water.State(**items) for items in last_two_frames)
    )
    return ModelState(state)


@gin.register
class PrimitiveEquationStateEncoder(hk.Module):
  """Encoder that extracts primitive_equations.State from inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      time_axis: int = 0,
      transform_module: TransformModule = EncoderIdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.slice_fn = functools.partial(
        pytree_utils.slice_along_axis, axis=time_axis, idx=-1)
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features, input_coords
    )

  def __call__(
      self, inputs: DataState, forcing: Forcing
  ) -> primitive_equations.State:
    del forcing
    state = self.transform_fn(
        primitive_equations.State(**self.slice_fn(inputs)))
    return ModelState(state)


@gin.register
class PrimitiveEquationLeapfrogEncoder(hk.Module):
  """Encoder that extracts primitive_equations.State pair from inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      time_axis: int = 0,
      transform_module: TransformModule = EncoderIdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.slice_fn = functools.partial(
        pytree_utils.slice_along_axis, axis=time_axis, idx=slice(-2, None))
    self.time_axis = time_axis
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features, input_coords
    )

  def __call__(
      self, inputs: DataState, forcing: Forcing
  ) -> Tuple[primitive_equations.State, ...]:
    del forcing
    last_two_frames = pytree_utils.split_axis(
        self.slice_fn(inputs), self.time_axis
    )
    state = self.transform_fn(
        tuple(primitive_equations.State(**items) for items in last_two_frames)
    )
    return ModelState(state)


@gin.register
class PrimitiveEquationStateWithTimeEncoder(hk.Module):
  """Encoder that extracts primitive_equations.StateWithTime from inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      time_axis: int = 0,
      transform_module: TransformModule = EncoderIdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.slice_fn = functools.partial(
        pytree_utils.slice_along_axis, axis=time_axis, idx=-1)
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features, input_coords
    )

  def __call__(
      self, inputs: DataState, forcing: Forcing
  ) -> ModelState:
    del forcing
    sliced_inputs = self.slice_fn(inputs)
    state = self.transform_fn(
        primitive_equations.StateWithTime(**sliced_inputs))
    return ModelState(state)


@gin.register
class WeatherbenchToPrimitiveEncoder(hk.Module):
  """Encoder that extracts primitive_equations.StateWithTime from WB inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      time_axis: int = 0,
      orography_module: OrographyModule = orographies.ClippedOrography,
      transform_module: TransformModule = EncoderIdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    ref_temps = aux_features[xarray_utils.REF_TEMP_KEY]
    self.ref_temps = ref_temps[..., np.newaxis, np.newaxis]
    self.coords = coords
    self.input_coords = input_coords
    self.slice_fn = functools.partial(
        pytree_utils.slice_along_axis, axis=time_axis, idx=-1)
    modal_orography_init_fn = orography_module(
        coords, dt, physics_specs, aux_features
    )
    modal_orography = modal_orography_init_fn()  # pytype: disable=not-callable  # jax-ndarray
    self.surface_pressure_fn = functools.partial(
        vertical_interpolation.get_surface_pressure,
        input_coords.vertical,
        orography=input_coords.horizontal.to_nodal(modal_orography),
        gravity_acceleration=physics_specs.gravity_acceleration,
    )
    self.curl_and_div_fn = functools.partial(
        spherical_harmonic.uv_nodal_to_vor_div_modal,
        input_coords.horizontal,
    )
    self.modal_interpolate_fn = coordinate_systems.get_spectral_interpolate_fn(
        input_coords, coords, expect_same_vertical=False
    )
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features, input_coords
    )

  def weatherbench_to_primitive(
      self,
      wb_state_nodal: weatherbench_utils.State,
  ) -> ModelState:
    """Converts wb_state on pressure coordinates to primitive on sigma."""
    # Note: the returned values have mixed nodal/modal representations.
    surface_pressure = self.surface_pressure_fn(wb_state_nodal.z)
    interpolate_fn = vertical_interpolation.vectorize_vertical_interpolation(
        vertical_interpolation.vertical_interpolation
    )
    regrid_fn = functools.partial(
        vertical_interpolation.interp_pressure_to_sigma,
        pressure_coords=self.input_coords.vertical,
        sigma_coords=self.coords.vertical,
        surface_pressure=surface_pressure,
        interpolate_fn=interpolate_fn,
    )
    wb_state_on_sigma = regrid_fn(wb_state_nodal)
    u, v = self.coords.physics_to_dycore_sharding(
        (wb_state_on_sigma.u, wb_state_on_sigma.v)
    )
    vorticity, divergence = self.coords.dycore_to_physics_sharding(
        self.curl_and_div_fn(u, v)
    )
    pe_state_on_sigma = primitive_equations.StateWithTime(
        divergence=divergence,
        vorticity=vorticity,
        temperature_variation=(wb_state_on_sigma.t - self.ref_temps),
        log_surface_pressure=jnp.log(surface_pressure),
        sim_time=wb_state_on_sigma.sim_time,
        tracers=wb_state_on_sigma.tracers,
    )
    return pe_state_on_sigma

  def __call__(
      self,
      inputs: DataState,
      forcing: Forcing,
  ) -> ModelState:
    del forcing
    wb_state = weatherbench_utils.State(**self.slice_fn(inputs))
    wb_state = coordinate_systems.maybe_to_nodal(wb_state, self.input_coords)
    pe_state = self.weatherbench_to_primitive(wb_state)
    pe_state = coordinate_systems.maybe_to_modal(pe_state, self.input_coords)
    pe_state = self.modal_interpolate_fn(pe_state)
    return ModelState(state=self.transform_fn(pe_state))


@gin.register
class LearnedWeatherbenchToPrimitiveEncoder(WeatherbenchToPrimitiveEncoder):
  """Same as `WeatherbenchToPrimitiveEncoder`, but with learned corrections."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      modal_to_nodal_data_features_module: FeaturesModule,
      modal_to_nodal_model_features_module: FeaturesModule,
      nodal_mapping_module: MappingModule,
      correction_transform_module: TransformModule,
      prediction_mask: typing.Pytree,
      time_axis: int = 0,
      orography_module: OrographyModule = orographies.ClippedOrography,
      transform_module: TransformModule = EncoderIdentityTransform,
      randomness_module: RandomnessModule = stochastic.ZerosRandomField,
      perturbation_module: PerturbationModule = perturbations.NoPerturbation,
      name: Optional[str] = None,
  ):
    super().__init__(
        coords,
        dt,
        physics_specs,
        aux_features,
        input_coords=input_coords,
        time_axis=time_axis,
        orography_module=orography_module,
        name=name,
    )
    self.prediction_mask = prediction_mask
    # data features are computed in real space on input coordinates.
    self.data_features_fn = modal_to_nodal_data_features_module(
        input_coords, dt, physics_specs, aux_features
    )
    self.model_features_fn = modal_to_nodal_model_features_module(
        coords, dt, physics_specs, aux_features
    )
    self.nodal_mapping_module = nodal_mapping_module
    self.output_transform_fn = correction_transform_module(
        input_coords, dt, physics_specs, aux_features
    )
    self.get_nodal_shape_fn = lambda x: coordinate_systems.get_nodal_shapes(
        x, coords
    )
    self.transform_fn = transform_module(
        coords, dt, physics_specs, aux_features, input_coords
    )
    self.randomness_fn = randomness_module(
        coords, dt, physics_specs, aux_features
    )
    self.perturbation_fn = perturbation_module(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      inputs: DataState,
      forcing: Forcing,
  ) -> ModelState:
    randomness = self.randomness_fn.unconditional_sample(
        hk.maybe_next_rng_key()
    )
    wb_state = self.coords.with_physics_sharding(
        weatherbench_utils.State(**self.slice_fn(inputs))
    )
    wb_state_nodal = self.coords.with_physics_sharding(
        coordinate_systems.maybe_to_nodal(wb_state, self.input_coords)
    )
    wb_state_modal = self.coords.with_physics_sharding(
        coordinate_systems.maybe_to_modal(wb_state, self.input_coords)
    )
    pe_state = self.coords.physics_to_dycore_sharding(
        self.weatherbench_to_primitive(wb_state_nodal)
    )
    # Computing corrections to the primitive_equations state.
    pe_state_modal = coordinate_systems.maybe_to_modal(
        pe_state, self.input_coords
    )
    # we need to interpolate `pe_state_modal` to self.coords to compute
    # features in model space. In most cases this is no-op as grids match.
    pe_state_modal = self.modal_interpolate_fn(pe_state_modal)
    pe_state_nodal = coordinate_systems.maybe_to_nodal(
        pe_state_modal, self.coords
    )
    prediction_shapes = jax.tree_util.tree_map(
        lambda x, y: self.get_nodal_shape_fn(x) if y else None,
        pe_state_nodal.asdict(),
        self.prediction_mask,
    )
    prediction_shapes = primitive_equations.StateWithTime(**prediction_shapes)
    net = self.nodal_mapping_module(prediction_shapes)
    # we need modal values to compute features for ML corrections.
    data_features = self.data_features_fn(
        wb_state_modal.asdict(), forcing=forcing,
    )
    model_features = self.model_features_fn(
        pe_state_modal.asdict(), forcing=forcing,
        randomness=randomness.nodal_value,
    )
    data_features = transforms.add_prefix(data_features, 'data_')
    model_features = transforms.add_prefix(model_features, 'model_')

    all_features = self.coords.with_physics_sharding(
        data_features | model_features
    )

    nodal_corrections = self.coords.with_physics_sharding(
        self.output_transform_fn(net(all_features))
    )

    perturbed_correction = self.perturbation_fn(
        state=None,  # Unused
        inputs=nodal_corrections,
        randomness=randomness.nodal_value,
    )

    add_fn = lambda x, y: x + y if y is not None else x
    corrected_pe_state = self.coords.physics_to_dycore_sharding(
        jax.tree_util.tree_map(
            add_fn,
            coordinate_systems.maybe_to_modal(pe_state_nodal, self.coords),
            coordinate_systems.maybe_to_modal(
                perturbed_correction, self.coords
            ),
        )
    )
    return ModelState(state=self.transform_fn(corrected_pe_state))


@gin.register
class DimensionalWeatherbenchToPrimitiveEncoder(WeatherbenchToPrimitiveEncoder):
  """Same as WeatherbenchToPrimitiveEncoder, but with dimensional inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      inputs_to_units_mapping: Dict[str, str],
      time_axis: int = 0,
      orography_module: OrographyModule = orographies.ClippedOrography,
      transform_module: TransformModule = EncoderIdentityTransform,
      name: Optional[str] = None,
  ):
    nondim_pressure_centers = physics_specs.nondimensionalize(
        input_coords.vertical.centers * scales.units.millibar
    )
    nondim_input_coords = coordinate_systems.CoordinateSystem(
        input_coords.horizontal,
        vertical_interpolation.PressureCoordinates(nondim_pressure_centers),
        spmd_mesh=input_coords.spmd_mesh,
    )
    super().__init__(
        coords,
        dt,
        physics_specs,
        aux_features,
        input_coords=nondim_input_coords,
        time_axis=time_axis,
        orography_module=orography_module,
        transform_module=transform_module,
        name=name,
    )
    self.nondim_transform_fn = transforms.NondimensionalizeTransform(
        coords,
        dt,
        physics_specs,
        aux_features,
        nondim_input_coords,
        inputs_to_units_mapping=inputs_to_units_mapping,
    )

  def __call__(
      self,
      inputs: DataState,
      forcing: Forcing,
  ) -> primitive_equations.StateWithTime:
    nondim_inputs = self.nondim_transform_fn(inputs)
    return super().__call__(nondim_inputs, forcing)


@gin.register
class DimensionalLearnedWeatherbenchToPrimitiveEncoder(
    LearnedWeatherbenchToPrimitiveEncoder
):
  """Same as LearnedWeatherbenchToPrimitiveEncoder, but with dimensional inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      inputs_to_units_mapping: Dict[str, str],
      modal_to_nodal_data_features_module: FeaturesModule,
      modal_to_nodal_model_features_module: FeaturesModule,
      nodal_mapping_module: MappingModule,
      correction_transform_module: TransformModule,
      prediction_mask: typing.Pytree,
      time_axis: int = 0,
      orography_module: OrographyModule = orographies.ClippedOrography,
      transform_module: TransformModule = EncoderIdentityTransform,
      randomness_module: RandomnessModule = stochastic.ZerosRandomField,
      perturbation_module: PerturbationModule = perturbations.NoPerturbation,
      name: Optional[str] = None,
  ):
    nondim_pressure_centers = physics_specs.nondimensionalize(
        input_coords.vertical.centers * scales.units.millibar
    )
    nondim_input_coords = coordinate_systems.CoordinateSystem(
        input_coords.horizontal,
        vertical_interpolation.PressureCoordinates(nondim_pressure_centers),
        spmd_mesh=input_coords.spmd_mesh,
    )
    super().__init__(
        coords,
        dt,
        physics_specs,
        aux_features,
        input_coords=nondim_input_coords,
        modal_to_nodal_data_features_module=modal_to_nodal_data_features_module,
        modal_to_nodal_model_features_module=(
            modal_to_nodal_model_features_module
        ),
        nodal_mapping_module=nodal_mapping_module,
        correction_transform_module=correction_transform_module,
        prediction_mask=prediction_mask,
        time_axis=time_axis,
        orography_module=orography_module,
        transform_module=transform_module,
        randomness_module=randomness_module,
        perturbation_module=perturbation_module,
        name=name,
    )
    self.nondim_transform_fn = transforms.NondimensionalizeTransform(
        coords,
        dt,
        physics_specs,
        aux_features,
        nondim_input_coords,
        inputs_to_units_mapping=inputs_to_units_mapping,
    )

  def __call__(
      self,
      inputs: DataState,
      forcing: Forcing,
  ) -> ModelState:
    nondim_inputs = self.nondim_transform_fn(inputs)
    return super().__call__(nondim_inputs, forcing)


@gin.register
class DimensionalLearnedWeatherbenchToPrimitiveWithMemoryEncoder(hk.Module):
  """Same as DimensionalLearnedWeatherbenchToPrimitiveEncoder, but with memory.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: Dict[str, Any],
      input_coords: coordinate_systems.CoordinateSystem,
      inputs_to_units_mapping: Dict[str, str],
      modal_to_nodal_data_features_module: FeaturesModule,
      modal_to_nodal_model_features_module: FeaturesModule,
      nodal_mapping_module: MappingModule,
      correction_transform_module: TransformModule,
      prediction_mask: typing.Pytree,
      time_axis: int = 0,
      orography_module: OrographyModule = orographies.ClippedOrography,
      transform_module: TransformModule = EncoderIdentityTransform,
      randomness_module: RandomnessModule = stochastic.ZerosRandomField,
      perturbation_module: PerturbationModule = perturbations.NoPerturbation,
      name: Optional[str] = None,
  ):
    nondim_pressure_centers = physics_specs.nondimensionalize(
        input_coords.vertical.centers * scales.units.millibar)
    nondim_input_coords = coordinate_systems.CoordinateSystem(
        input_coords.horizontal,
        vertical_interpolation.PressureCoordinates(nondim_pressure_centers),
        spmd_mesh=input_coords.spmd_mesh,
    )
    super().__init__(name=name)
    make_encoder_fn = functools.partial(
        LearnedWeatherbenchToPrimitiveEncoder,
        coords=coords, dt=dt,
        physics_specs=physics_specs, aux_features=aux_features,
        input_coords=nondim_input_coords,
        modal_to_nodal_data_features_module=
        modal_to_nodal_data_features_module,
        modal_to_nodal_model_features_module=
        modal_to_nodal_model_features_module,
        nodal_mapping_module=nodal_mapping_module,
        correction_transform_module=correction_transform_module,
        prediction_mask=prediction_mask, time_axis=time_axis,
        orography_module=orography_module,
        transform_module=transform_module,
        name=name
    )

    # Memory will be deterministic. State may be random.
    self.memory_encoder = make_encoder_fn(
        randomness_module=stochastic.NoRandomField,
        perturbation_module=perturbations.NoPerturbation,
    )
    self.state_encoder = make_encoder_fn(
        randomness_module=randomness_module,
        perturbation_module=perturbation_module,
    )

    self.nondim_transform_fn = transforms.NondimensionalizeTransform(
        coords, dt, physics_specs, aux_features, nondim_input_coords,
        inputs_to_units_mapping=inputs_to_units_mapping)

  def __call__(
      self,
      inputs: DataState,
      forcing: Forcing,
  ) -> ModelState:
    nondim_inputs = self.nondim_transform_fn(inputs)
    memory = self.memory_encoder(nondim_inputs, forcing=forcing)
    model_state = self.state_encoder(nondim_inputs, forcing=forcing)
    return ModelState(
        state=model_state.state,
        memory=memory.state,
        randomness=model_state.randomness,
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of custom initializers for NN parameters."""

from typing import Any, Optional, Sequence

import gin
import haiku as hk
import jax
import numpy as np


# Registering default initializers.
Constant = gin.external_configurable(hk.initializers.Constant)
VarianceScaling = gin.external_configurable(hk.initializers.VarianceScaling)
Orthogonal = gin.external_configurable(hk.initializers.Orthogonal)


def _compute_fans(
    shape: Sequence[int],
    fan_in_axes: Optional[Sequence[int]] = None,
) -> tuple[int, int]:
  """Computes the number of input and output units for a weight shape."""
  # adapted from dm-haiku/_src/initializers.py
  if len(shape) < 1:
    fan_in = fan_out = 1
  elif len(shape) == 1:
    fan_in = fan_out = shape[0]
  elif len(shape) == 2:
    fan_in, fan_out = shape
  else:
    if fan_in_axes is not None:
      # Compute fan-in using user-specified fan-in axes.
      fan_in = np.prod([shape[i] for i in fan_in_axes])
      fan_out = np.prod([s for i, s in enumerate(shape)
                         if i not in fan_in_axes])
    else:
      # If no axes specified, assume convolution kernels (2D, 3D, or more.)
      # kernel_shape: (..., input_depth, depth)
      receptive_field_size = np.prod(shape[:-2])
      fan_in = shape[-2] * receptive_field_size
      fan_out = shape[-1] * receptive_field_size
  return fan_in, fan_out


@gin.register
class ReducingVarianceScaling(hk.initializers.Initializer):
  """Initializer that result in variance that reduces as width increases.

  Initializes weights that result in features with expected variance of
  `scale / n`, where `n` corresponds to the width of the layer. This initializer
  can be used in the output layer to achieve µ parameterization [1].

  References:
    [1]: https://arxiv.org/abs/2203.03466
  """

  def __init__(
      self,
      scale=1.0,
      mode='fan_in',
      distribution='truncated_normal',
      fan_in_axes=None,
  ):
    """Constructs `ReducingVarianceScaling` initializer.

    Args:
      scale: Variance scale for a width == 1 initialization.
      mode: One of ``fan_in``, ``fan_out``, ``fan_avg``
      distribution: Random distribution to use. One of ``truncated_normal``,
        ``normal`` or ``uniform``.
      fan_in_axes: Optional sequence of int specifying which axes of the shape
        are part of the fan-in. If none provided, then the weight is assumed
        to be like a convolution kernel, where all leading dimensions are part
        of the fan-in, and only the trailing dimension is part of the fan-out.
        Useful if instantiating multi-headed attention weights.
    """
    if scale < 0.0:
      raise ValueError('`scale` must be a positive float.')
    if mode not in {'fan_in', 'fan_out', 'fan_avg'}:
      raise ValueError('Invalid `mode` argument:', mode)
    distribution = distribution.lower()
    if distribution not in {'normal', 'truncated_normal', 'uniform'}:
      raise ValueError('Invalid `distribution` argument:', distribution)
    self.scale = scale
    self.mode = mode
    self.distribution = distribution
    self.fan_in_axes = fan_in_axes

  def __call__(self, shape: Sequence[int], dtype: Any) -> jax.Array:
    scale = self.scale
    fan_in, fan_out = _compute_fans(shape, self.fan_in_axes)
    if self.mode == 'fan_in':
      scale /= max(1.0, fan_in) ** 2
    elif self.mode == 'fan_out':
      scale /= max(1.0, fan_out) ** 2
    else:
      scale /= max(1.0, (fan_in + fan_out) / 2.0) ** 2

    if self.distribution == 'truncated_normal':
      stddev = np.sqrt(scale)
      # Adjust stddev for truncation.
      # Constant from scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
      distribution_stddev = np.asarray(.87962566103423978, dtype=dtype)
      stddev = stddev / distribution_stddev
      return hk.initializers.TruncatedNormal(stddev=stddev)(shape, dtype)
    elif self.distribution == 'normal':
      stddev = np.sqrt(scale)
      return hk.initializers.RandomNormal(stddev=stddev)(shape, dtype)
    else:
      limit = np.sqrt(3.0 * scale)
      uniform_init = hk.initializers.RandomUniform(minval=-limit, maxval=limit)
      return uniform_init(shape, dtype)



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""API for providing dynamic inputs to NeuralGCM models."""

import abc
import functools

from flax import nnx
import jax
import jax.numpy as jnp

from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import typing


class DynamicInputValue(nnx.Intermediate):
  ...


class DynamicInputModule(nnx.Module, abc.ABC):
  """Base class for modules that interface with dynamically supplied data."""

  @abc.abstractmethod
  def update_dynamic_inputs(self, dynamic_inputs):
    """Ingests relevant data from `dynamic_inputs` onto the internal state."""
    raise NotImplementedError()

  @abc.abstractmethod
  def output_shapes(self) -> typing.Pytree:
    raise NotImplementedError()

  @abc.abstractmethod
  def __call__(self, sim_time: float) -> typing.Pytree:
    """Returns dynamic data at the specified sim_time."""
    raise NotImplementedError()


class DynamicInputSlice(DynamicInputModule):
  """Exposes inputs from the most recent available time slice."""

  def __init__(
      self,
      keys_to_coords: dict[str, cx.Coordinate],
      time_axis: int = 0,
  ):
    self.keys_to_coords = keys_to_coords
    self.time_axis = time_axis

  def update_dynamic_inputs(self, dynamic_inputs):
    # TODO(dkochkov): check that data aligns with expected data_specs.
    self.times = DynamicInputValue({
        k: dynamic_inputs[k].timestamp
        for k in self.keys_to_coords.keys()
    })
    self.data = DynamicInputValue(
        {k: dynamic_inputs[k].field.data for k in self.keys_to_coords.keys()}
    )

  def output_shapes(self) -> typing.Pytree:
    return {
        k: typing.ShapeFloatStruct(v.shape)
        for k, v in self.keys_to_coords.items()
    }

  def __call__(
      self,
      sim_time: float,
  ) -> typing.Pytree:
    """Returns covariates at the specified sim_time."""
    outputs = {}
    for k, time in self.times.value.items():  # pylint: disable=attribute-error
      time_indices = jnp.arange(time.size)
      approx_index = jnp.interp(sim_time, time, time_indices)
      index = jnp.round(approx_index).astype('int32')
      field_index_fn = functools.partial(
          jax.lax.dynamic_index_in_dim,
          index=index,
          axis=self.time_axis,
          keepdims=False,
      )
      sliced_data = field_index_fn(self.data.value[k])
      outputs[k] = cx.wrap(sliced_data, self.keys_to_coords[k])
    return outputs



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for towers.py."""
import functools
from typing import Callable

from absl.testing import absltest
from absl.testing import parameterized
from flax import nnx
import jax
import jax.numpy as jnp
from neuralgcm.experimental import standard_layers
from neuralgcm.experimental import towers
import numpy as np


class TowersTest(parameterized.TestCase):
  """Tests Tower primitives."""

  @parameterized.named_parameters(
      dict(
          testcase_name='_mlp_uniform',
          nn_input_shape=(12,),
          spatial_shape=(10, 20),
          output_size=4,
          nn_factory=functools.partial(
              standard_layers.MlpUniform,
              hidden_size=3,
              n_hidden_layers=4,
              use_bias=True,
          ),
      ),
      dict(
          testcase_name='_mlp_uniform_with_remat',
          nn_input_shape=(3,),
          spatial_shape=(16, 7),
          output_size=3,
          nn_factory=functools.partial(
              standard_layers.MlpUniform,
              hidden_size=1,
              n_hidden_layers=13,
              use_bias=False,
          ),
          apply_remat=True,
      ),
      dict(
          testcase_name='_cnn',
          nn_input_shape=(4, 12),  # here 12 corresponds to level axis.
          spatial_shape=(6, 6),
          output_size=2,
          nn_factory=functools.partial(
              standard_layers.CnnLevel,
              channels=(3, 5, 7),
              kernel_sizes=3,
              use_bias=False,
          ),
      ),
  )
  def test_column_tower(
      self,
      nn_input_shape: tuple[int, ...],
      spatial_shape: tuple[int, ...],
      output_size: int,
      nn_factory: standard_layers.UnaryLayerFactory,
      apply_remat: bool = False,
  ):
    input_size = nn_input_shape[0]
    tower = towers.ColumnTower(
        input_size,
        output_size,
        column_net_factory=nn_factory,
        rngs=nnx.Rngs(0),
        apply_remat=apply_remat,
    )
    inputs = jax.random.uniform(
        jax.random.PRNGKey(42), nn_input_shape + spatial_shape
    )
    outputs = tower(inputs)
    with self.subTest('output_shape'):
      expected_shape = (output_size,) + nn_input_shape[1:] + spatial_shape
      self.assertEqual(outputs.shape, expected_shape)
    with self.subTest('same_params_count_as_column'):
      net = nn_factory(input_size, output_size, rngs=nnx.Rngs(1))
      expected = sum(
          [np.prod(x.shape) for x in jax.tree.leaves(nnx.state(net, nnx.Param))]
      )
      actual = sum([
          np.prod(x.shape) for x in jax.tree.leaves(nnx.state(tower, nnx.Param))
      ])
      self.assertEqual(actual, expected)

  @parameterized.named_parameters(
      dict(
          testcase_name='_mlp_uniform',
          nn_input_shape=(12,),
          spatial_shape=(10, 20),
          output_size=4,
          tower_factory=functools.partial(
              towers.ColumnTower,
              column_net_factory=functools.partial(
                  standard_layers.MlpUniform, hidden_size=3, n_hidden_layers=4
              ),
          ),
      ),
      dict(
          testcase_name='_cnn',
          nn_input_shape=(4, 12),  # here 12 corresponds to level axis.
          spatial_shape=(6, 6),
          output_size=2,
          tower_factory=functools.partial(
              towers.ColumnTower,
              column_net_factory=functools.partial(
                  standard_layers.CnnLevel,
                  channels=(3, 5, 7),
                  kernel_sizes=3,
                  use_bias=False,
              ),
          ),
      ),
  )
  def test_epd_tower(
      self,
      nn_input_shape: tuple[int, ...],
      spatial_shape: tuple[int, ...],
      output_size: int,
      tower_factory: towers.UnaryTowerFactory,
  ):
    input_size = nn_input_shape[0]
    tower = towers.EpdTower(
        input_size,
        output_size,
        latent_size=8,
        num_process_blocks=2,
        encode_tower_factory=tower_factory,
        process_tower_factory=tower_factory,
        decode_tower_factory=tower_factory,
        rngs=nnx.Rngs(0),
    )
    inputs = jax.random.uniform(
        jax.random.PRNGKey(42), nn_input_shape + spatial_shape
    )
    outputs = tower(inputs)
    with self.subTest('output_shape'):
      expected_shape = (output_size,) + nn_input_shape[1:] + spatial_shape
      self.assertEqual(outputs.shape, expected_shape)

  @parameterized.parameters(
      dict(
          input_size=1,
          output_size=1,
          kernel_size=(5, 5),
          num_hidden_units=12,
          num_hidden_layers=2,
          dilation=1,
          activation=jax.nn.relu,
          activate_final=True,
          use_bias=True,
          apply_remat=False,
      ),
      dict(
          input_size=2,
          output_size=1,
          kernel_size=(3, 3),
          num_hidden_units=24,
          num_hidden_layers=2,
          dilation=4,
          activation=jax.nn.gelu,
          activate_final=True,
          use_bias=False,
          apply_remat=False,
      ),
      dict(
          input_size=6,
          output_size=5,
          kernel_size=(3, 3),
          num_hidden_units=12,
          num_hidden_layers=0,
          dilation=2,
          activation=jax.nn.gelu,
          activate_final=False,
          use_bias=True,
          apply_remat=True,
      ),
  )
  def test_conv_lon_lat_tower(
      self,
      input_size: int,
      output_size: int,
      kernel_size: tuple[int, int],
      num_hidden_units: int,
      num_hidden_layers: int,
      dilation: int,
      activation: Callable[[jnp.ndarray], jnp.ndarray],
      activate_final: bool,
      use_bias: bool,
      apply_remat: bool,
  ):
    conv_tower = towers.ConvLonLatTower(
        input_size=input_size,
        output_size=output_size,
        kernel_size=kernel_size,
        num_hidden_units=num_hidden_units,
        num_hidden_layers=num_hidden_layers,
        dilation=dilation,
        activation=activation,
        activate_final=activate_final,
        use_bias=use_bias,
        apply_remat=apply_remat,
        rngs=nnx.Rngs(0),
    )

    params = nnx.state(conv_tower, nnx.Param)

    with self.subTest('output_shape'):
      inputs = jnp.ones((input_size, 56, 45))
      outputs = conv_tower(inputs)
      self.assertEqual(outputs.shape, (output_size,) + inputs.shape[1:])

    with self.subTest('total_params_count'):
      expected = count_conv_tower_params(
          input_size,
          output_size,
          kernel_size,
          num_hidden_units,
          num_hidden_layers,
          use_bias,
      )
      actual = sum([np.prod(x.shape) for x in jax.tree.leaves(params)])
      self.assertEqual(actual, expected)

  @parameterized.parameters(
      dict(
          input_size=1,
          output_size=1,
          hidden_sizes=(1, 1),
          conv_block_hidden_layers=2,
          use_bias=True,
          kernel_size=(5, 5),
          dilations=(1, 2, 4),
          activation=jax.nn.gelu,
          apply_remat=False,
          num_hidden_layers_skip_residual=0,
      ),
      dict(
          input_size=1,
          output_size=1,
          hidden_sizes=(1, 1),
          conv_block_hidden_layers=2,
          use_bias=True,
          kernel_size=(5, 5),
          dilations=1,
          activation=jax.nn.relu,
          apply_remat=False,
          num_hidden_layers_skip_residual=0,
      ),
      dict(
          input_size=3,
          output_size=1,
          hidden_sizes=(1, 2, 9),
          conv_block_hidden_layers=2,
          use_bias=True,
          kernel_size=(3, 3),
          dilations=(1, 2, 4, 1),
          activation=jax.nn.tanh,
          apply_remat=False,
          num_hidden_layers_skip_residual=1,
      ),
      dict(
          input_size=10,
          output_size=5,
          hidden_sizes=(1, 1),
          conv_block_hidden_layers=2,
          use_bias=True,
          kernel_size=(3, 3),
          dilations=(1, 2, 1),
          activation=jax.nn.relu,
          apply_remat=True,
          num_hidden_layers_skip_residual=2,
      ),
  )
  def test_resnet(
      self,
      input_size: int,
      output_size: int,
      hidden_sizes: tuple[int, int],
      conv_block_hidden_layers: int,
      use_bias: bool,
      kernel_size: tuple[int, int],
      dilations: tuple[int, ...] | int,
      activation: Callable[[jnp.ndarray], jnp.ndarray],
      apply_remat: bool,
      num_hidden_layers_skip_residual: int,
  ):

    resnet_tower = towers.ResNet(
        input_size=input_size,
        output_size=output_size,
        conv_block_num_channels=hidden_sizes,
        conv_block_num_hidden_layers=conv_block_hidden_layers,
        use_bias=use_bias,
        kernel_size=kernel_size,
        dilations=dilations,
        activation=activation,
        num_hidden_layers_skip_residual=num_hidden_layers_skip_residual,
        apply_remat=apply_remat,
        rngs=nnx.Rngs(0),
    )
    with self.subTest('output_shape'):
      inputs = jnp.ones((input_size, 128, 64))
      outputs = resnet_tower(inputs)
      self.assertEqual(outputs.shape, (output_size,) + inputs.shape[1:])
      inputs = jnp.ones((input_size, 127, 61))
      outputs = resnet_tower(inputs)
      self.assertEqual(outputs.shape, (output_size,) + inputs.shape[1:])


def count_conv_tower_params(
    input_size: int,
    output_size: int,
    kernel_size: tuple[int, int],
    num_hidden_units: int,
    num_hidden_layers: int,
    use_bias: bool,
):

  def _count_conv_params(input_size, kernel_size, output_size, use_bias):
    n_w_params = np.prod(kernel_size) * input_size * output_size
    n_b_params = output_size
    return n_w_params + n_b_params * use_bias

  param_count = 0
  param_count += _count_conv_params(
      input_size, kernel_size, num_hidden_units, use_bias
  )
  for _ in range(num_hidden_layers):
    param_count += _count_conv_params(
        num_hidden_units, kernel_size, num_hidden_units, use_bias
    )
  param_count += _count_conv_params(
      num_hidden_units, kernel_size, output_size, use_bias
  )
  return param_count


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()
